#!/usr/bin/env python
"""
Shuffled-Complex-Evolution (SCE) algorithm for function minimization

This code is based on a Fortran program of Qingyun Duan (2004), ported to
Python by Stijn Van Hoey (2011). It was taken up, debugged, enhanced and is
maintained by Matthias Cuntz while at Department of Computational Hydrosystems
(CHS), Helmholtz Centre for Environmental Research - UFZ, Leipzig, Germany, and
continued while at Institut National de Recherche pour l'Agriculture,
l'Alimentation et l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2004-2022 Qingyun Duan [1]_, Stijn Van Hoey, Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

References
----------
.. [1] Duan, Sorooshian and Gupta (1992) Effective and efficient global
       optimization for conceptual rainfall-runoff models, Water Resour Res 28,
       1015-1031, https://doi.org/10.1029/91WR02985

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   sce

History
    * Written in Fortran by Q Duan, Sep 2004
    * Ported to Python by Stijn Van Hoey, 2011
      https://github.com/stijnvanhoey/Optimization_SCE
    * Synchronised with enhanced Fortran version of CHS,
      Oct 2013, Matthias Cuntz
    * Added functionality to call external executable, Nov 2016, Matthias Cuntz
    * Treat NaN and Inf in function output, Nov 2016, Matthias Cuntz
    * Possibility to exclude (mask) parameters from optimisation,
      Nov 2016, Matthias Cuntz
    * Added restart possibility, Nov 2016, Matthias Cuntz
    * Return also function value of best parameter set if maxit==True,
      Nov 2016, Matthias Cuntz
    * Removed functionality to call external executable,
      Dec 2017, Matthias Cuntz
    * Print out number of function evaluations with printit=1,
      Mar 2018, Matthias Cuntz
    * Mask parameters with degenerated ranges, e.g. upper<lower bound,
      Mar 2018, Matthias Cuntz
    * Use only masked parameters in calculation of geometric range,
      Mar 2018, Matthias Cuntz
    * Removed bug that calculated the size of the complexes using all
      parameters and not only the masked parameters, Mar 2018, Matthias Cuntz
    * Fixed bug where masked parameters were always out of bounds,
      Mar 2018, Matthias Cuntz
    * Allow scalar bounds, which will be taken for all parameters,
      Mar 2018, Matthias Cuntz
    * Get number of parameters from lower boundary in SampleInputMatrix,
      Mar 2018, Matthias Cuntz
    * Define number of parameters by inquiring x0 in case of restart,
      May 2018, Matthias Cuntz
    * Removed multiplication with one hundred in criter_change,
      regarded a bug compared to Fortran code, May 2018, Matthias Cuntz
    * Removed exec command to make restart work with Python 3,
      May 2020, Matthias Cuntz
    * Use underscore before private routines, May 2020, Matthias Cuntz
    * Code refactoring, Sep 2021, Matthias Cuntz
    * Added keywords args and kwargs to pass to function,
      Apr 2022, Matthias Cuntz
    * Pass RandomState to _SampleInputMatrix, Jul 2022, Matthias Cuntz
    * Different sampling of input parameters, Jul 2022, Matthias Cuntz
    * Use helper class to pass arguments to objective function,
      Dec 2022, Matthias Cuntz
    * Copy strtobool from distutils.util because distutils is deprecated,
      Dec 2022, Matthias Cuntz
    * Include number of complexes ngs in restartfile, Dec 2022, Matthias Cuntz
    * No restart files written by default, Dec 2022, Matthias Cuntz
    * Rewrite into class SCESolver, Dec 2022, Matthias Cuntz
    * Output OptimizeResult class, Dec 2022, Matthias Cuntz
    * Polish results with L-BFGS-B, Dec 2022, Matthias Cuntz

"""
import warnings
import numpy as np
from scipy.optimize import OptimizeResult, minimize
from scipy._lib._util import check_random_state
from scipy.optimize._constraints import Bounds

# ToDo:
# - write tmp/population files (of Fortran code)

__all__ = ['sce']


# extended scipy/_util.py version
class _FunctionWrapper:
    """
    Wrap user function with arguments and keywords, allowing picklability

    Parameters
    ----------
    func : callable
        Function in the form ``func(x, *args, **kwargs)``, where ``x`` are
        the parameters in the form of an iterable.
        ``args`` and ``kwargs`` are passed to the function via the usual
        unpacking operators.

    """
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.nargs = len(args)
        self.nkwargs = len(kwargs)
        if self.nkwargs == 0:
            if ( (len(self.args) == 2) and
                 isinstance(self.args[-1], dict) and
                 (len(self.args[-1]) == 0) ):
                # if kwargs={} then **kwargs={} and hence counted as args
                self.args = self.args[0]
                self.nargs = len(args)

    def __call__(self, x):
        if (self.nargs > 0) and (self.nkwargs > 0):
            return self.func(x, *self.args, **self.kwargs)
        elif (self.nargs > 0) and (self.nkwargs == 0):
            return self.func(x, *self.args)
        elif (self.nargs == 0) and (self.nkwargs > 0):
            return self.func(x, **self.kwargs)
        else:
            return self.func(x)


# from distutils.util (Python 3.11.1)
def _strtobool(val):
    """
    Convert a string representation of truth to true (1) or false (0).
    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.

    """
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return 1
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return 0
    else:
        raise ValueError("invalid truth value {!r}".format(val))


def sce(func, x0, lb, ub=None,
        mask=None,
        args=(), kwargs={},
        sampling='half-open',
        maxn=1000, kstop=10, pcento=0.0001, peps=0.001,
        ngs=2, npg=0, nps=0, nspl=0, mings=0,
        seed=None, iniflg=True,
        alpha=0.8, beta=0.45, maxit=False, printit=2,
        polish=True,
        restart=False, restartfile1='',
        restartfile2=''):
    """
    Shuffled-Complex-Evolution algorithm for function minimization

    The SCE or SCE-UA method is a general purpose global optimization.
    The algorithm has been described in detail in Duan et al. [1]_ and [2]_.
    Another paper of Duan et al. [3]_ discusses how to use the method
    effectively.
    This implementation also includes the recommendations of Behrangi et al.
    [4]_.

    Parameters
    ----------
    func : callable
        Function in the form ``func(x, *args, **kwargs)``, where ``x`` are
        the parameters in the form of an iterable.
        ``args`` and ``kwargs`` are passed to the function via the usual
        unpacking operators.
    x0 : array_like
        Parameter values with `mask==1` used in initial complex if
        `iniflg==1`.
    lb : array_like, sequence or `Bounds`
        Lower bounds of parameters if ``ub`` given.
        If `ub` is not given, then `lb` are the bounds of the parameters,
        either 1) as an instance of the `Bounds` class, or 2) as ``(min, max)``
        pairs for each element in ``x``, defining the finite lower and upper
        bounds for the parameters of `func`.
    ub : array_like, optional
        Upper bounds of parameters.
    mask : array_like, optional
        Include (1, True) or exclude (0, False) parameters in minimization
        (default: include all parameters). The number of parameters ``nopt`` is
        ``sum(mask)``.
    args : tuple, optional
        Extra arguments passed to the function *func*. Note that ``args`` must
        be iterable. `args=int`and `args=(int)` are not valid (with int being
        any scalar variable) but should be `args=(int,)`.
    kwargs : dict, optional
        Extra keyword arguments passed to the function `func`.
    sampling : string or array_like of strings, optional
        Options for sampling random numbers. Options can be on of:

            - 'half-open': same as 'right-half-open' [lb, ub)
            - 'left-half-open': sample random floats in half-open
              interval (lb, ub]
            - 'right-half-open': sample random floats in half-open
              interval [lb, ub)
            - 'open': sample random floats in open interval (lb, ub)
            - 'log': sample half-open interval [log(lb), log(ub)), which
              samples better intervals spanning orders or magnitude such as
              `lb=1e-9` and `ub=1e-4`

        The default is 'half-open'.
    maxn : int, optional
        Maximum number of function evaluations allowed during minimization
        (without polishing) (default: 1000).
    kstop : int, optional
        Maximum number of evolution loops before convergence (default: 10).
    pcento : float, optional
        Percentage change allowed in kstop loops before convergence
        (default: 0.0001).
    peps : float, optional
        Value of normalised geometric range needed for convergence
        (default: 0.001).
    ngs : int, optional
        Number of complexes (default: 2).
    npg : int, optional
        Number of points in each complex (default: `2*nopt+1`).
    nps : int, optional
        Number of points in each sub-complex (default: `nopt+1`).
    mings : int, optional
        Minimum number of complexes required if the number of complexes is
        allowed to reduce as the optimization proceeds (default: `ngs`).
    nspl : int, optional
        Number of evolution steps allowed for each complex before complex
        shuffling (default: `2*nopt+1`).
    seed : {None, int, `numpy.random.Generator`, `numpy.random.RandomState`}, optional
        If `seed` is None (or `numpy.random`), the `numpy.random.RandomState`
        singleton is used.
        If `seed` is an int, a new `RandomState` instance is used,
        seeded with `seed`.
        If `seed` is already a `Generator` or `RandomState` instance then
        that instance is used.
        Specify `seed` for repeatable results.
    iniflg : bool, optional
        If True: include initial parameters ``x0`` in initial population
        (default: True).
    alpha : float, optional
        Parameter for reflection of points in complex (default: 0.8).
    beta : float, optional
        Parameter for contraction of points in complex (default: 0.45).
    maxit : bool, optional
        If True: maximize instead of minimize `func` (default: False).
    printit : int, optional
        Controlling print-out (default: 2):

            - 0: print information for the best point of the population
            - 1: print information for each function evaluation
            - 2: no printing.

        The default is 2.
    polish : bool, optional
        If True (default), then `scipy.optimize.minimize` is used with the
        `L-BFGS-B` method to polish the result at the end, which
        can improve the minimization slightly. For large problems, polishing
        can take a long time due to the computation of the Jacobian.
    restart : bool, optional
        If True, continue from saved state in `restartfile1` and
        `restartfile2` (default: False).
    restartfile1 : str, optional
        Filename for saving state of array variables of `SCE` (default: '').
        If `restart==True` and `restartfile1==''` then
        `restartfile1='sce.restart.npz'` will be taken.
    restartfile2 : int, optional
        Filename for saving state of non-array variables of `SCE`
        (default: `restartfile1 + '.txt'`).

    Returns
    -------
    res : OptimizeResult
        The optimization result represented as a `OptimizeResult` object.
        Important attributes are: ``x`` the solution array, ``success`` a
        Boolean flag indicating if the optimizer exited successfully (0) and
        ``message`` which describes the cause of the termination. See
        `OptimizeResult` for a description of other attributes. If `polish`
        was employed, and a lower minimum was obtained by the polishing, then
        OptimizeResult also contains the ``jac`` attribute.

    References
    ----------
    .. [1] Duan QY, Sorooshian S, and Gupta VK,
           Effective and efficient global optimization for conceptual
           rainfall-runoff models, Water Resources Research 28(4), 1015-1031,
           1992, https://doi.org/10.1029/91WR02985
    .. [2] Duan QY, Gupta VK, and Sorooshian S,
           Shuffled Complex Evolution approach for effective and efficient
           global minimization, Journal of Optimization Theory and its
           Applications 76(3), 501-521, 1993,
           https://doi.org/10.1007/BF00939380
    .. [3] Duan QY, Gupta VK, and Sorooshian S,
           Optimal use of the SCE-UA global optimization method for calibrating
           watershed models, Journal of Hydrology 158, 265-284, 1994,
           https://doi.org/10.1016/0022-1694(94)90057-4
    .. [4] Behrangi A, Khakbaz B, Vrugt JA, Duan Q, and Sorooshian S
           Comment on "Dynamically dimensioned search algorithm for
           computationally efficient watershed model calibration" by
           Bryan A. Tolson and Christine A. Shoemaker, Water Resources
           Research 44, W12603, 2008, http://doi.org/10.1029/2007WR006429

    Examples
    --------
    Search the minimum of the Rosenbrock function, implemented in `rosen`
    in `scipy.optimize`.

    >>> import numpy as np
    >>> from scipy.optimize import rosen

    One has to provide the function `rosen`, an initial guess of the
    parameters, and the lower and upper limits of the parameters.
    The 2D version is:

    >>> lb = np.array([-5., -2.])
    >>> ub = np.array([5., 8.])
    >>> x0 = np.array([-2., 7.])
    >>> res = sce(rosen, x0, lb, ub, seed=1, maxn=1000, printit=2)
    >>> print(res.nfev)
    298
    >>> print('{:.3f}'.format(res.fun))
    0.001

    A 10-dimensional version using `(min. max)` pars for the bounds and
    setting a number of parameters could be:

    >>> nopt = 10
    >>> lb = np.full(10, -5.)
    >>> ub = np.full(10, 5.)
    >>> x0 = np.full(10, 0.5)
    >>> res = sce(rosen, x0, zip(lb, ub), maxn=30000, kstop=10, pcento=0.0001,
    ...     seed=12358, ngs=5, npg=5*nopt+1, nps=nopt+1, nspl=5*nopt+1,
    ...     mings=2, iniflg=True, printit=2, alpha=0.8, beta=0.45)
    >>> print(res.nfev)
    30228
    >>> print('{:.3g}'.format(res.fun))
    3.38e-12

    """
    # using a context manager means that any created Pool objects are
    # cleared up.
    with SCESolver(func, x0, lb, ub=ub,
                   mask=mask, args=args, kwargs=kwargs, sampling=sampling,
                   maxn=maxn, kstop=kstop, pcento=pcento,
                   ngs=ngs, npg=npg, nps=nps, nspl=nspl, mings=mings,
                   peps=peps, seed=seed, iniflg=iniflg,
                   alpha=alpha, beta=beta, maxit=maxit, printit=printit,
                   polish=polish,
                   restart=restart, restartfile1=restartfile1,
                   restartfile2=restartfile2) as solver:
        ret = solver.solve()

    return ret


class SCESolver:
    """
    This class implements the Shuffled-Complex-Evolution algorithm

    Parameters
    ----------
    func : callable
        Function in the form ``func(x, *args, **kwargs)``, where ``x`` are
        the parameters in the form of an iterable.
        ``args`` and ``kwargs`` are passed to the function via the usual
        unpacking operators.
    x0 : array_like
        Parameter values with `mask==1` used in initial complex if
        `iniflg==1`.
    lb : array_like, sequence or `Bounds`
        Lower bounds of parameters if ``ub`` given.
        If `ub` is not given, then `lb` are the bounds of the parameters,
        either 1) as an instance of the `Bounds` class, or 2) as ``(min, max)``
        pairs for each element in ``x``, defining the finite lower and upper
        bounds for the parameters of `func`.
    ub : array_like
        Upper bounds of parameters.
    mask : array_like, optional
        Include (1, True) or exclude (0, False) parameters in minimization
        (default: include all parameters). The number of parameters ``nopt`` is
        ``sum(mask)``.
    args : tuple, optional
        Extra arguments passed to the function *func*.
    kwargs : dict, optional
        Extra keyword arguments passed to the function `func`.
    sampling : string or array_like of strings, optional
        Options for sampling random numbers. Options can be on of:

            - 'half-open': same as 'right-half-open' [lb, ub)
            - 'left-half-open': sample random floats in half-open
              interval (lb, ub]
            - 'right-half-open': sample random floats in half-open
              interval [lb, ub)
            - 'open': sample random floats in open interval (lb, ub)
            - 'log': sample half-open interval [log(lb), log(ub)), which
              samples better intervals spanning orders or magnitude such as
              `lb=1e-9` and `ub=1e-4`

        The default is 'half-open'.
    maxn : int, optional
        Maximum number of function evaluations allowed during minimization
        (without polishing) (default: 1000).
    kstop : int, optional
        Maximum number of evolution loops before convergence (default: 10).
    pcento : float, optional
        Percentage change allowed in kstop loops before convergence
        (default: 0.0001).
    peps : float, optional
        Value of normalised geometric range needed for convergence
        (default: 0.001).
    ngs : int, optional
        Number of complexes (default: 2).
    npg : int, optional
        Number of points in each complex (default: `2*nopt+1`).
    nps : int, optional
        Number of points in each sub-complex (default: `nopt+1`).
    mings : int, optional
        Minimum number of complexes required if the number of complexes is
        allowed to reduce as the optimization proceeds (default: `ngs`).
    nspl : int, optional
        Number of evolution steps allowed for each complex before complex
        shuffling (default: `2*nopt+1`).
    seed : {None, int, `numpy.random.Generator`, `numpy.random.RandomState`}, optional
        If `seed` is None (or `numpy.random`), the `numpy.random.RandomState`
        singleton is used.
        If `seed` is an int, a new `RandomState` instance is used,
        seeded with `seed`.
        If `seed` is already a `Generator` or `RandomState` instance then
        that instance is used.
        Specify `seed` for repeatable results.
    iniflg : bool, optional
        If True: include initial parameters ``x0`` in initial population
        (default: True).
    alpha : float, optional
        Parameter for reflection of points in complex (default: 0.8).
    beta : float, optional
        Parameter for contraction of points in complex (default: 0.45).
    maxit : bool, optional
        If True: maximize instead of minimize `func` (default: False).
    printit : int, optional
        Controlling print-out (default: 2):

            - 0: print information for the best point of the population
            - 1: print information for each function evaluation
            - 2: no printing.

        The default is 2.
    polish : bool, optional
        If True (default), then `scipy.optimize.minimize` is used with the
        `L-BFGS-B` method to polish the result at the end, which
        can improve the minimization slightly. For large problems, polishing
        can take a long time due to the computation of the Jacobian.
    restart : bool, optional
        If True, continue from saved state in `restartfile1` and
        `restartfile2` (default: False).
    restartfile1 : str, optional
        Filename for saving state of array variables of `SCE` (default: '').
        If `restart==True` and `restartfile1==''` then
        `restartfile1='sce.restart.npz'` will be taken.
    restartfile2 : int, optional
        Filename for saving state of non-array variables of `SCE`
        (default: `restartfile1 + '.txt'`).

    """

    def __init__(self, func, x0, lb, ub=None,
                 mask=None, args=(), kwargs={},
                 sampling='half-open',
                 maxn=1000, kstop=10, pcento=0.0001,
                 ngs=2, npg=0, nps=0, nspl=0, mings=0,
                 peps=0.001, seed=None, iniflg=True,
                 alpha=0.8, beta=0.45, maxit=False, printit=2,
                 polish=True,
                 restart=False, restartfile1='',
                 restartfile2=''):

        # function to minimize
        self.func = _FunctionWrapper(func, *args, **kwargs)
        # seed random number generator - will be overwritten by restart
        # self.rnd = np.random.RandomState(seed=seed)
        self.rnd = check_random_state(seed)
        # parameters for initial run and for restart
        self.sampling = sampling
        self.maxn = maxn
        self.kstop = kstop
        self.pcento = pcento
        self.peps = peps
        self.alpha = alpha
        self.beta = beta
        self.maxit = maxit
        self.printit = printit
        self.polish = polish
        self.restartfile1 = restartfile1
        self.restartfile2 = restartfile2
        if restart and (not self.restartfile1):
            self.restartfile1 = 'sce.restart.npz'
        if self.restartfile1 and (not self.restartfile2):
            self.restartfile2 = self.restartfile1 + '.txt'
        if not restart:
            # initialize SCE parameters
            self.nn    = len(x0)
            self.mask  = np.ones(self.nn, dtype=bool) if mask is None else mask
            self.nopt  = np.sum(self.mask)
            self.ngs   = ngs if ngs > 0 else 2
            self.npg   = npg if npg > 0 else 2 * self.nopt + 1
            self.nps   = nps if nps > 0 else self.nopt + 1
            self.nspl  = nspl if nspl > 0 else 2 * self.nopt + 1
            self.mings = mings if mings > 0 else self.ngs
            self.npt   = self.npg * self.ngs

            # assure lb and ub are numpy arrays
            if ub is None:
                if isinstance(lb, Bounds):
                    self.lb = lb.lb
                    self.ub = lb.ub
                else:
                    self.lb, self.ub = zip(*lb)
            else:
                self.lb = lb
                self.ub = ub
            self.lb = np.array(self.lb)
            self.ub = np.array(self.ub)

            # same bounds for all parameters
            try:
                if len(self.lb) == 1:
                    self.lb = np.full(self.nn, self.lb[0])
            except TypeError:  # could be size 0 array if lb was a scalar
                self.lb = np.full(self.nn, self.lb)
            try:
                if len(self.ub) == 1:
                    self.ub = np.full(self.nn, self.ub[0])
            except TypeError:
                self.ub = np.full(self.nn, self.ub)
            bound = self.ub - self.lb
            # degenerated bounds
            if np.any(bound <= 0.):
                ii = np.where(bound <= 0.)[0]
                warnings.warn(
                    f'sce: found lower bound lb >= upper bound lb for'
                    f' parameter(s) {ii} with lb={self.lb[ii]} and'
                    f' ub={self.ub[ii]} => masking the parameter(s).',
                    UserWarning, stacklevel=2)
                self.mask[ii] = False
                self.ub[ii] = self.lb[ii]

            # set 'large' for function runs that return NaN
            self.large = 0.5 * np.finfo(float).max

            # create an initial population to fill array x(npt, nparams)
            self.x = self.sample_input_matrix(self.npt)
            for i in range(self.npt):
                self.x[i, :] = np.where(self.mask, self.x[i, :], x0)
            if iniflg == 1:
                self.x[0, :] = x0

            self.icall = 0
            self.xf = np.zeros(self.npt)
            for i in range(self.npt):
                fuc = self.func(self.x[i, :])
                self.xf[i] = -fuc if self.maxit else fuc
                self.icall += 1
                if self.printit == 1:
                    print('  i, f, X: ', self.icall, self.xf[i], self.x[i, :])

            # remember large for treating of NaNs
            self.large = self.xf[np.isfinite(self.xf)].max()
            self.large = (1.1 * self.large if self.large > 0. else
                          0.9 * self.large)

            # sort the population in order of increasing function values
            # and report best point
            self.xf = np.where(np.isfinite(self.xf), self.xf, self.large)
            idx = np.argsort(self.xf)
            self.xf = self.xf[idx]
            self.x  = self.x[idx, :]
            self.bestx  = self.x[0, :]
            self.bestf  = self.xf[0]

            # compute the normalized geometric range of the parameters
            self.gnrng = self.calc_gnrng()

            # initialise evolution loops
            self.nloop = 0
            self.criter = []
            self.criter_change = 1e+5

            # save restart
            self.write_restartfiles()

        else:  # if not restart
            self.nn = len(x0)
            self.read_restartfiles()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return self

    # def __iter__(self):
    #     return self

    def __next__(self):
        # loop on complexes (sub-populations)
        for igs in range(self.ngs):
            k1 = np.array(range(self.npg))
            k2 = k1 * self.ngs + igs

            # partition the population into complexes (sub-populations)
            cx = np.zeros((self.npg, self.nn))
            cf = np.zeros((self.npg))
            k1 = np.array(range(self.npg))
            k2 = k1 * self.ngs + igs
            cx[k1, :] = self.x[k2, :]
            cf[k1]    = self.xf[k2]

            # evolve sub-population igs for nspl steps:
            for loop in range(self.nspl):
                # select simplex by sampling the complex according to a
                # linear probability distribution
                lcs = np.zeros(self.nps, dtype=int)
                lcs[0] = 1
                for k3 in range(1, self.nps):
                    for i in range(1000):
                        lpos = int(np.floor(
                            self.npg + 0.5 -
                            np.sqrt((self.npg + 0.5)**2 -
                                    self.npg * (self.npg + 1) *
                                    self.rnd.random_sample(1)) ))
                        # check if element was already chosen
                        idx = (lcs[0:k3] == lpos).nonzero()
                        if idx[0].size == 0:
                            break
                    lcs[k3] = lpos
                lcs.sort()

                # construct the simplex
                s  = np.zeros((self.nps, self.nn))
                s  = cx[lcs, :]
                sf = cf[lcs]

                # remember large for treating of NaNs
                self.large = cf[np.isfinite(cf)].max()
                self.large = (1.1 * self.large if self.large > 0. else
                              0.9 * self.large)

                snew, fnew, icall = self.cce(s, sf)
                # replace the worst point in simplex with the new point
                s[-1, :] = snew
                sf[-1]   = fnew
                # self.icall += icall
                self.icall += icall

                # reinsert the simplex into the complex
                cx[lcs, :] = s
                cf[lcs]    = sf

                # sort the complex
                cf  = np.where(np.isfinite(cf), cf, self.large)
                idx = np.argsort(cf)
                cf  = cf[idx]
                cx  = cx[idx, :]
                # end of inner loop for competitive evolution of simplexes:
                #     for loop in range(self.nspl):
                # i.e. end of evolve sub-population igs for nspl steps

            self.x[k2, :] = cx[k1, :]
            self.xf[k2]   = cf[k1]

        return

    def calc_criter_change(self):
        """
        Computes the percentage change in function output

        """
        if self.nloop >= self.kstop:
            criter_change = np.abs(self.criter[self.nloop - 1] -
                                   self.criter[self.nloop - self.kstop])
            criter_change = (
                criter_change /
                np.maximum( 1e-15, np.mean( np.abs(
                    self.criter[self.nloop - self.kstop:self.nloop]) ) ) )
        else:
            criter_change = self.criter_change

        return criter_change

    def calc_gnrng(self):
        """
        Computes the normalized geometric range of the parameters

        """
        bound = self.ub - self.lb
        rrange = (np.ma.array(self.x.max(axis=0) - self.x.min(axis=0),
                              mask=~self.mask) /
                  np.ma.array(bound, mask=~self.mask))
        gnrng = np.ma.exp(np.ma.mean(np.ma.log(rrange)))

        return gnrng

    def cce(self, s, sf):
        """
        Generate a new point in a simplex

        Parameters
        ----------
        s : 2D-array
            The sorted simplex in order of increasing function values
        sf : 1D-array
            Function values in increasing order

        Results
        -------
        new parameter set, function value of new set,
        number of function evaluations performed

        """
        # best point
        sb = s[0, :]
        # fb = sf[0]
        # worst point and function value
        sw = s[-1, :]
        fw = sf[-1]

        # centroid of the simplex excluding worst point
        ce = np.mean(s[:-1, :], axis=0)

        # attempt a reflection point
        snew = ce + self.alpha * (ce - sw)
        # sb should have initial params at mask==False
        snew = np.where(self.mask, snew, sb)

        # check if new point is outside bounds
        ibound = 0
        if np.ma.any(np.ma.array(snew - self.lb, mask=~self.mask) < 0.):
            ibound = 1
        if np.ma.any(np.ma.array(self.ub - snew, mask=~self.mask) < 0.):
            ibound = 2
        if ibound >= 1:
            snew = self.sample_input_matrix(1)[0, :]
            snew = np.where(self.mask, snew, sb)

        icall = 0
        # calc function for reflection point
        fuc = self.func(snew)
        fnew = -fuc if self.maxit else fuc
        icall += 1
        if self.printit == 1:
            print('  i, f, X: ', self.icall + icall, fnew, snew)

        # reflection failed: attempt a contraction point
        if fnew > fw:
            snew = sw + self.beta * (ce - sw)
            snew = np.where(self.mask, snew, sb)
            fuc = self.func(snew)
            fnew = -fuc if self.maxit else fuc
            icall += 1
            if self.printit == 1:
                print('  i, f, X: ', self.icall + icall, fnew, snew)

        # both reflection and contraction have failed, attempt a random point
        if fnew > fw:
            snew = self.sample_input_matrix(1)[0, :]
            snew = np.where(self.mask, snew, sb)
            fuc = self.func(snew)
            fnew = -fuc if self.maxit else fuc
            icall += 1
            if self.printit == 1:
                print('  i, f, X: ', self.icall + icall, fnew, snew)

        # end of cce
        return snew, fnew, icall

    def check_number_calls(self):
        """
        Check for maximum number of function calls reached

        """
        if self.icall < self.maxn:
            return True
        else:
            if self.printit < 2:
                print(f'Optimisation terminated because trial number'
                      f' {self.maxn} reached maximum number of trials'
                      f' {self.icall}.')
            return False

    def check_criter_change(self):
        """
        Check if percentage change in function output is below pcento

        """
        if self.criter_change < self.pcento:
            if self.printit < 2:
                print(f'The best point has improved by less then {self.pcento}'
                      f' in the last {self.kstop} loops.')
            return True
        else:
            return False

    def check_geometric_range(self):
        """
        Check if normalized geometric range of the parameters is below peps

        """
        self.gnrng = self.calc_gnrng()
        if self.gnrng < self.peps:
            if self.printit < 2:
                print(f'The population has converged to a small parameter'
                      f' space {self.gnrng} (<{self.peps}).')
            return True
        else:
            return False

    def read_restartfiles(self):
        """
        Read from restart files

        """
        with open(self.restartfile1, 'rb') as p1:
            pp = np.load(p1)
            self.lb       = pp['lb']
            self.ub       = pp['ub']
            self.mask     = pp['mask']
            self.criter   = pp['criter']
            self.x        = pp['x']
            self.xf       = pp['xf']
            rs2           = pp['rs2']

        with open(self.restartfile2, 'r') as p2:
            (self.ngs, self.nopt, self.npg, self.nps, self.nspl,
             self.mings, self.npt, self.nloop, self.icall, rs3, rs4) = [
                 int(inn) for inn in p2.readline().rstrip().split(',') ]
            (self.gnrng, self.criter_change, self.bestf, rs5) = [
                float(inn) for inn in p2.readline().rstrip().split(',') ]
            self.maxit = bool(_strtobool(p2.readline().rstrip()))
            rs1 = p2.readline().rstrip()

        self.rnd.set_state((rs1, rs2, rs3, rs4, rs5))

    def write_restartfiles(self):
        """
        Write restart files if demanded

        """
        if self.restartfile1:
            # compute the normalized geometric range of the parameters
            self.gnrng = self.calc_gnrng()
            # compute percentage change in function output
            self.criter_change = self.calc_criter_change()
            # get the current state of the random number generator
            rs1, rs2, rs3, rs4, rs5 = self.rnd.get_state()

            # only arrays into savez_compressed - restartfile1
            np.savez_compressed(self.restartfile1, lb=self.lb, ub=self.ub,
                                mask=self.mask,
                                criter=self.criter, x=self.x, xf=self.xf,
                                bestx=self.bestx, rs2=rs2)

            # scalars into text file - restartfile2
            with open(self.restartfile2, 'w') as p:
                iout = [self.ngs, self.nopt, self.npg, self.nps, self.nspl,
                        self.mings, self.npt, self.nloop, self.icall,
                        rs3, rs4]
                iform = '{:d},' * (len(iout) - 1) + '{:d}'
                print(iform.format(*iout), file=p)
                fout = [self.gnrng, self.criter_change, self.bestf, rs5]
                fform = '{:.15g},' * (len(fout) - 1) + '{:.15g}'
                print(fform.format(*fout), file=p)
                print(self.maxit, file=p)
                print(rs1, file=p)

        return

    def sample_input_matrix(self, npt=0):
        """
        Create input parameter matrix (npt, npars) for
        npt simulations and npars parameters with bounds lb and ub

        Returns
        -------
        parameter matrix (npt, npars) with new parameter samples

        """
        if npt < 1:
            npt = self.npt
        npars = len(self.lb)
        if isinstance(self.sampling, str):
            isampling = [self.sampling] * npars
        else:
            isampling = self.sampling
        assert len(isampling) == npars, (
            f'sampling must be string or list of strings'
            f' with {npars} entries')
        x = np.zeros((npt, npars))
        bound = self.ub - self.lb
        for i in range(npt):
            irnd = self.rnd.random_sample(npars)
            for j in range(npars):
                opt = isampling[j].lower()
                if (opt == 'half-open') or (opt == 'right-half-open'):
                    x[i, j] = self.lb[j] + irnd[j] * bound[j]
                elif opt == 'left-half-open':
                    irnd[j] = 1. - irnd[j]
                    x[i, j] = self.lb[j] + irnd[j] * bound[j]
                elif opt == 'open':
                    iirnd = irnd[j]
                    while not (iirnd > 0.):
                        iirnd = self.rnd.random_sample(1)
                    x[i, j] = self.lb[j] + iirnd * bound[j]
                elif opt == 'log':
                    # x must be > 0. for ln(x)
                    xshift = 0.
                    if (self.lb[j] * self.ub[j]) < 0.:
                        # lb < 0 and ub > 0 -> shift both > 0
                        xshift = 2. * np.maximum(np.abs(self.lb[j]),
                                                 np.abs(self.ub[j]))
                    elif (self.lb[j] * self.ub[j]) == 0.:
                        if self.lb[j] == 0.:
                            # lb == 0 and ub > 0 -> shift to [ub, 2*ub)
                            xshift = self.ub[j]
                        if self.ub[j] == 0.:
                            # lb < 0 and ub == 0 -> shift to [-lb, -2*lb) > 0.
                            xshift = -2. * self.lb[j]
                    elif (self.lb[j] < 0.) and (self.ub[j] < 0.):
                        # lb < 0 and ub < 0 -> shift both > 0
                        xshift = 2. * np.maximum(np.abs(self.lb[j]),
                                                 np.abs(self.ub[j]))
                    lnlb = np.log(self.lb[j] + xshift)
                    lnub = np.log(self.ub[j] + xshift)
                    x[i, j] = np.exp(lnlb + irnd[j] * (lnub - lnlb)) - xshift
                else:
                    raise ValueError(f'unknown sampling option'
                                     f' {isampling[j]}.\n'
                                     f'Known samplings are: half-open,'
                                     f' left-half-open, right-half-open, open,'
                                     f' log')

        return x

    def solve(self):
        while (self.check_number_calls() and
               (not self.check_geometric_range()) and
               (not self.check_criter_change())):
            self.nloop += 1

            # loop on complexes (sub-populations)
            next(self)

            # shuffle the complexes
            idx = np.argsort(self.xf)
            self.xf = self.xf[idx]
            self.x  = self.x[idx, :]

            # record best and worst points
            self.bestx = self.x[0, :]
            self.bestf = self.xf[0]

            if self.printit < 2:
                print(f'\nEvolution loop {self.nloop}, trials {self.icall}.'
                      f' Best f: {self.bestf:.2f}, worst f:'
                      f' {self.xf[-1]:.2f}\n'
                      f'  best x:\n', self.bestx, '\n')

            self.criter = np.append(self.criter, self.bestf)
            self.criter_change = self.calc_criter_change()

            # save restart
            self.write_restartfiles()

        # finish up
        if self.printit < 2:
            print('Search stopped at trial number {0:d} with normalized'
                  ' geometric range {1:f}. '.format(self.icall, self.gnrng))
            print('The best point has improved by {:f} in the last'
                  ' {:d} loops.'.format(self.criter_change, self.kstop))

        if not self.check_number_calls():
            success = False
            status  = 2
            message = f'Reached maximum number of trials {self.maxn}'
        if self.check_criter_change():
            success = True
            status  = 1
            message = (f'Best point improved less than {self.pcento}'
                       f' in last {self.kstop} loops')
        if self.check_geometric_range():
            success = True
            status  = 0
            message = (f'Normalized geometric range of parameters'
                       f' {self.gnrng} < {self.peps}')

        if self.maxit:
            self.bestf *= -1.

        sce_result = OptimizeResult(
            x=self.bestx,
            success=success,
            status=status,
            message=message,
            fun=self.bestf,
            nfev=self.icall,
            nit=self.nloop)

        # polish results, only works if no mask
        if self.polish and (len(self.lb) == self.nopt):
            polish_method = 'L-BFGS-B'
            result = minimize(self.func,
                              np.copy(sce_result.x),
                              method=polish_method,
                              bounds=zip(self.lb, self.ub))

            self.icall += result.nfev
            sce_result.nfev = self.icall

            # Polishing solution is only accepted if there is an improvement in
            # the cost function, the polishing was successful and the solution
            # lies within the bounds.
            if ( (result.fun < sce_result.fun) and
                 result.success and
                 np.all(result.x <= self.ub) and
                 np.all(self.lb <= result.x) ):
                sce_result.fun = result.fun
                sce_result.x = result.x
                sce_result.jac = result.jac

        return sce_result


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # from pyjams.functions import ackley, griewank, goldstein_price
    # from pyjams.functions import rastrigin, rosenbrock, six_hump_camelback
    # from scipy.optimize import Bounds
    # maxn = 10000
    # seed = 12345

    # """
    # This is the Ackley Function
    # Global Optimum (n>=2): 0.0 at origin
    # """
    # npara = 10
    # lb = -10 * np.ones(npara)
    # ub = 10 * np.ones(npara)
    # x0 = np.random.rand(npara) * 10.
    # res = sce(ackley, x0, lb, ub, maxn=maxn,
    #           restartfile1='', seed=seed, polish=False)
    # print('Ackley unpolished ', res.x, res.fun)
    # # print('       ', res.message)
    # res = sce(ackley, x0, lb, ub, maxn=maxn,
    #           restartfile1='', seed=seed)
    # print('Ackley ', res.x, res.fun)
    # # print('       ', res.message)

    # """
    # This is the Griewank Function (2-D or 10-D)
    # Bound: X(i)=[-600,600], for i=1,2,...,10  !for visualization only 2!
    #    Global Optimum: 0, at origin
    # """
    # npara = 10
    # lb = -600 * np.ones(npara)
    # ub = 600 * np.ones(npara)
    # x0 = np.random.rand(npara) * 600.
    # res = sce(griewank, x0, lb, ub, maxn=maxn,
    #           restartfile1='', seed=seed)
    # print('Griewank ', res.x, res.fun)
    # # print('       ', res.message)

    # """
    # This is the Goldstein-Price Function
    # Bound X1=[-2,2], X2=[-2,2]
    # Global Optimum: 3.0,(0.0,-1.0)
    # """
    # npara = 2
    # lb = -2 * np.ones(npara)
    # ub = 2 * np.ones(npara)
    # x0 = np.random.rand(npara) * 2.
    # res = sce(goldstein_price, x0, zip(lb, ub), maxn=maxn,
    #           restartfile1='', seed=seed)
    # print('Goldstein ', res.x, res.fun)
    # # print('       ', res.message)

    # """
    # This is the Rastrigin Function
    # Bound: X1=[-1,1], X2=[-1,1]
    # Global Optimum: -2, (0,0)
    # """
    # npara = 2
    # lb = -1 * np.ones(npara)
    # ub = 1 * np.ones(npara)
    # x0 = np.random.rand(npara) * 1.
    # bounds = Bounds(lb, ub)
    # res = sce(rastrigin, x0, bounds, maxn=maxn,
    #           restartfile1='', seed=seed)
    # print('Rastrigin ', res.x, res.fun)
    # # print('       ', res.message)

    # """
    # This is the Rosenbrock Function
    # Bound: X1=[-5,5], X2=[-2,8]; Global Optimum: 0,(1,1)
    #        lb=[-5 -5]; ub=[5 5]; x0=[1 1];
    # """
    # npara = 2
    # lb = -2 * np.ones(npara)
    # ub = 5 * np.ones(npara)
    # x0 = np.random.rand(npara) * 2.
    # res = sce(rosenbrock, x0, lb, ub, maxn=maxn,
    #           seed=seed)
    # print('Rosenbrock ', res.x, res.fun)
    # # print('       ', res.message)

    # """
    # This is the Six-hump Camelback Function.
    # Bound: X1=[-5,5], X2=[-5,5]
    # True Optima: -1.031628453489877, (-0.08983,0.7126), (0.08983,-0.7126)
    # """
    # npara = 2
    # lb = -5 * np.ones(npara)
    # ub = 5 * np.ones(npara)
    # x0 = np.random.rand(npara) * 5.
    # res = sce(six_hump_camelback, x0, lb, ub, maxn=maxn,
    #           seed=seed)
    # print('Six_hump_camelback ', res.x, res.fun)
    # # print('       ', res.message)

    # # Restart
    # maxn = 500
    # npara = 2
    # lb = -2 * np.ones(npara)
    # ub = 5 * np.ones(npara)
    # x0 = np.zeros(npara)
    # res = sce(rosenbrock, x0, lb, ub, maxn=maxn,
    #           restartfile1='',
    #           seed=seed, restart=False)
    # print('Rosenbrock Reference - ', res.x, res.fun)
    # # print('       ', res.message)
    # res = sce(rosenbrock, x0, lb, ub, maxn=maxn // 2,
    #           seed=seed, restart=False,
    #           restartfile1='sce.restart.npz')
    # print('Rosenbrock Restart 1 - ', res.x, res.fun)
    # res = sce(rosenbrock, x0, lb, ub, maxn=maxn,
    #           seed=seed, restart=True,
    #           restartfile1='sce.restart.npz')
    # print('Rosenbrock Restart 2 - ', res.x, res.fun)
    # # print('       ', res.message)
    # # Clean restart
    # import os
    # os.remove('sce.restart.npz')
    # os.remove('sce.restart.npz.txt')

    # # different sampling
    # npara = 2
    # lb = -2 * np.ones(npara)
    # ub = 5 * np.ones(npara)
    # x0 = np.random.rand(npara) * 2.
    # # sampling is str
    # for sampling in ['half-open', 'left-half-open',
    #                  'right-half-open', 'open', 'log']:
    #     res = sce(rosenbrock, x0, lb, ub, sampling=sampling,
    #               maxn=maxn, seed=seed)
    #     print('Rosenbrock0 ', sampling, res.x, res.fun)
    #     # print('       ', res.message)
    # # sampling is list
    # sampling = ['half-open', 'log']
    # res = sce(rosenbrock, x0, lb, ub, sampling=sampling,
    #           maxn=maxn, seed=seed)
    # print('Rosenbrock1 ', sampling, res.x, res.fun)
    # # print('       ', res.message)
    # # bounds include 0
    # lb = [0., -2.]
    # ub = [5., 8.]
    # x0 = np.random.rand(npara) * 2.
    # for sampling in ['half-open', 'left-half-open',
    #                  'right-half-open', 'open', 'log']:
    #     res = sce(rosenbrock, x0, lb, ub, sampling=sampling,
    #               maxn=maxn, seed=seed)
    #     print('Rosenbrock2 ', sampling, res.x, res.fun)
    #     # print('       ', res.message)
    # # bounds all > 0
    # lb = [0.1, 0.1]
    # ub = [5., 8.]
    # x0 = np.random.rand(npara) * 2.
    # for sampling in ['half-open', 'left-half-open',
    #                  'right-half-open', 'Open', 'log']:
    #     res = sce(rosenbrock, x0, lb, ub, sampling=sampling,
    #               maxn=maxn, restartfile1='', seed=seed)
    #     print('Rosenbrock3 ', sampling, res.x, res.fun)
    #     # print('       ', res.message)



#!/usr/bin/env python
"""
GUI dialogs to choose files and directories using Tkinter

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2015-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   directory_from_gui
   directories_from_gui
   file_from_gui
   files_from_gui

History
    * Written Jun 2014 by Matthias Cuntz (mc (at) macu (dot) de)
    * Added directories_from_gui, Oct 2015, Matthias Cuntz
    * Using numpy docstring format, May 2020, Matthias Cuntz
    * Port to pyjams, Jan 2022, Matthias Cuntz

"""


__all__ = ['directory_from_gui', 'directories_from_gui',
           'file_from_gui', 'files_from_gui']


# -------------------------------------------------------------------------
# Choose one directory in GUI
#

def directory_from_gui(
        initialdir='.',
        title='Choose directory'):  # pragma: no cover
    """
    Opens dialog to one select directory

    Parameters
    ----------
    initialdir : str, optional
        Initial directory, in which opens GUI (default: '.')
    title : str, optional
        Title of GUI (default: 'Choose directory')

    Returns
    -------
    str
        Selected directory

    Examples
    --------
    .. code-block:: python

       if not idir:
           idir = directory_from_gui()
           if not idir:
               raise ValueError('Error: no directory given.')

    """
    import tkinter as Tkinter
    import tkinter.filedialog as tkFileDialog

    root = Tkinter.Tk()
    root.withdraw()  # hide root window, i.e. white square

    # always on top
    # focus on (hidden) window so that child is on top
    root.tk.call('wm', 'attributes', '.', '-topmost', 1)

    direcs = tkFileDialog.askdirectory(
        parent=root, title=title, initialdir=initialdir)

    root.destroy()

    return direcs


# -------------------------------------------------------------------------
# Choose directories in GUI
#

def directories_from_gui(
        initialdir='.',
        title='Choose one or several directories'):  # pragma: no cover
    """
    Open dialog to select several directories

    Parameters
    ----------
    initialdir : str, optional
        Initial directory, in which opens GUI (default: '.')
    title : str, optional
        Title of GUI (default: 'Choose one or several directories')

    Returns
    -------
    list
        Selected directories

    Examples
    --------
    .. code-block:: python

       if not direcs:
           direcs = directories_from_gui()
           if not direcs:
               raise ValueError('Error: no directories given.')

    """
    import tkinter as Tkinter
    import tkinter.filedialog as tkFileDialog

    root = Tkinter.Tk()
    root.withdraw()  # hide root window, i.e. white square

    # always on top
    # focus on (hidden) window so that child is on top
    root.tk.call('wm', 'attributes', '.', '-topmost', 1)

    idir = initialdir
    alldirecs = []
    while True:
        direcs = tkFileDialog.askdirectory(
            parent=root, title=title, initialdir=idir)
        if not direcs:
            break
        alldirecs.append(direcs)
        idir = direcs

    root.destroy()

    return alldirecs


# -------------------------------------------------------------------------
# Choose one file in GUI
#

def file_from_gui(initialdir='.', title='Choose file',
                  multiple=False):  # pragma: no cover
    """
    Wrapper for :func:`files_from_gui` with multiple=False, i.e.
    open dialog to select one file

    Examples
    --------
    .. code-block:: python

       if not file:
           file = file_from_gui()
           if not file:
               raise ValueError('Error: no input file given.')

    """
    return files_from_gui(initialdir=initialdir,
                          title=title,
                          multiple=multiple)


# -------------------------------------------------------------------------
# Choose files in GUI
#

def files_from_gui(initialdir='.', title='Choose file(s)',
                   multiple=True):  # pragma: no cover
    """
    Open dialog to select one or several files

    Parameters
    ----------
    initialdir : str, optional
        Initial directory, in which opens GUI (default: '.')
    title : str, optional
        Title of GUI (default: 'Choose file(s)')
    multiple : bool, optional
        Allow selection of multiple files if True (default),
        else it is only possible to select one single file.

    Returns
    -------
    list
        Selected file(s)

    Notes
    -----
    It always returns a list even with `multiple=False`.

    Examples
    --------
    .. code-block:: python

       if not files:
           files = files_from_gui()
           if not files:
               raise ValueError('Error: no input file(s) given.')

    """
    import tkinter as Tkinter
    import tkinter.filedialog as tkFileDialog
    root = Tkinter.Tk()
    root.withdraw()  # hide root window, i.e. white square

    # always on top
    # focus on (hidden) window so that child is on top
    root.tk.call('wm', 'attributes', '.', '-topmost', 1)

    files = tkFileDialog.askopenfilename(
        parent=root, title=title, multiple=multiple, initialdir=initialdir)
    files = list(root.tk.splitlist(files))

    root.destroy()

    return files


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
This is the unittest for alpha_kin_h2o module.

python -m unittest -v tests/test_alpha_kin_h2o.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_alpha_kin_h2o.py

"""
import unittest


class TestAlphaKinH2O(unittest.TestCase):
    """
    Tests for alpha_kin_h2o.py
    """

    def test_alpha_kin_h2o(self):
        import numpy as np
        from pyjams import alpha_kin_h2o

        alpha = alpha_kin_h2o(isotope=0)
        self.assertEqual(np.around(alpha, 4), 1.0)

        epsilon = alpha_kin_h2o(isotope=1, eps=True) * 1000.
        self.assertEqual(np.around(epsilon, 4), 25.1153)

        epsilon = alpha_kin_h2o(isotope=2, eps=True, greater1=False) * 1000.
        self.assertEqual(np.around(epsilon, 4), -27.3)

        epsilon = alpha_kin_h2o(isotope=2, eps=True, boundary=True) * 1000.
        self.assertEqual(np.around(epsilon, 4), 18.6244)

        epsilon = alpha_kin_h2o(isotope=1, eps=True, cappa=True) * 1000.
        self.assertEqual(np.around(epsilon, 4), 16.3635)

        epsilon = alpha_kin_h2o(isotope=2, eps=True, greater1=False,
                                boundary=True, cappa=True) * 1000.
        self.assertEqual(np.around(epsilon, 4), -20.7076)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
Index in array which entry is closest to a given number.

This module was written by Matthias Cuntz while at Department of Computational
Hydrosystems, Helmholtz Centre for Environmental Research - UFZ, Leipzig,
Germany, and continued while at Institut National de Recherche pour
l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2012-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   closest

History
    * Written Jan 2012 by Matthias Cuntz (mc (at) macu (dot) de)
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Make numpy docstring format, Apr 2020, Matthias Cuntz
    * Ported into pyjams, Oct 2021, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""
from __future__ import division, absolute_import, print_function
import numpy as np


__all__ = ['closest']


def closest(arr, num, value=False):
    """
    Index in array which entry is closest to a given number.

    Closest returns the index of an array (arr) at which the entry is closest
    to a given number (num), which is `argmin(abs(arr-num))`.

    Parameters
    ----------
    arr : array_like
        Array to search closest entry
    num : number
        Number to which the closest entry is searched for in arr
    value : bool, optional
        Returns closest array element instead of index if True (default: False)

    Returns
    -------
    index : int
        Index of element closest to given number in flattend array.
        Use np.unravel_index to get index tuple.

    Examples
    --------
    >>> arr = np.arange(100)/99.*5.
    >>> print(closest(arr, 3.125))
    62
    >>> out = closest(arr, 3.125, value=True)
    >>> print('{:.3f}'.format(out))
    3.131

    >>> arr = np.arange(100).reshape((10,10))/99.*5.
    >>> out = closest(arr, 3.125, value=True)
    >>> print('{:.3f}'.format(out))
    3.131
    >>> print(closest(arr, 3.125))
    62
    >>> ii = np.unravel_index(closest(arr, 3.125), arr.shape)
    >>> print(ii)
    (6, 2)
    >>> out = arr[ii]
    >>> print('{:.3f}'.format(out))
    3.131

    """
    out = np.ma.argmin(np.ma.abs(np.ma.array(arr)-num))
    if value:
        return arr.flat[out]
    else:
        return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
This is the unittest for the means module.

python -m unittest -v tests/test_means.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_means.py

355: seasonalmonth
457-471: month
542-564: minute
581: meanday with one hour == undef
585-598: meanmonth
602-629: seasonalday
633-650: seasonalmeanday

"""
import unittest


def _flatten(itr):
    import numpy as np
    fitr = np.array(itr).flatten()
    if len(fitr) == 0:
        return list(fitr)
    else:
        if isinstance(fitr[0], str):
            return [ i for i in fitr ]
        else:
            return [ i if np.isfinite(i) else np.finfo(float).max
                     for i in fitr ]


class TestMeans(unittest.TestCase):
    """
    Tests for means.py
    """
    def setUp(self):
        import numpy as np
        from pyjams import num2date
        # 2 years at 00:15 and 00:45
        self.year0 = 1990
        self.nyears = 2
        self.ndays = 365
        self.nhours = 24
        self.nminutes = 2
        self.jdates = []
        for i in range(self.nyears):
            for j in range(self.ndays):
                for k in range(self.nhours):
                    for m in range(self.nminutes):
                        self.jdates.append(
                            self.year0 + i +
                            j / 365 +
                            k / (24 * 365) +
                            (15 + m * 30) /
                            (60 * 24 * 365))
        # to stay above second (rounding errors) when calculations with jdates
        muleps = np.finfo(float).eps
        self.jdates = [ dd + np.abs(dd) * muleps for dd in self.jdates ]
        self.ddates = num2date(self.jdates, calendar='decimal')
        self.ddates = [ dd.round_microseconds() for dd in self.ddates ]
        self.format = '%Y-%m-%d %H:%M:%S'
        self.dates = [ dd.strftime(self.format) for dd in self.ddates ]
        self.n = len(self.dates)
        self.x = np.arange(self.n) + 1
        self.undef = -9999.

    def test_means(self):
        import cftime as cf
        import datetime as pydatetime
        import numpy as np
        from pyjams import means
        from pyjams import date2num, num2date

        excelcalendars = ['excel', 'excel1900', 'excel1904']
        decimalcalendars = ['decimal', 'decimal365']
        calendars = excelcalendars + decimalcalendars
        # CF calendars are not working because of microsecond precision
        # cfcalendars = ['standard', 'gregorian', 'proleptic_gregorian',
        #                'noleap', 'julian', 'all_leap', '365_day', '366_day',
        #                '360_day']
        # calendars = cfcalendars
        ncal = len(calendars)
        itest = -1

        # different date formats

        dsoll = (min(self.jdates) + max(self.jdates)) * 0.5
        dsoll = num2date(dsoll, calendar='decimal', format=self.format)
        # dsoll = dsoll.round_microseconds()
        # dsoll = dsoll.strftime('%Y-%m-%d %H:%M:%S')
        n = self.nyears * self.ndays * self.nhours * self.nminutes
        xsoll = n//2 * (n + 1) / n
        print('')

        # dates = str
        itest += 1
        print('itest', itest)
        dates = self.dates  # str
        dout, xout = means(dates, self.x)
        assert isinstance(dout, type(dates[0]))
        assert dout == dsoll
        assert xout == xsoll

        # dates = str w/ format
        itest += 1
        print('itest', itest)
        iform = '%Y-%m-%d %H:%M'
        dates = [ dd[:16] for dd in self.dates ]
        dout, xout = means(dates, self.x, format=iform)
        assert isinstance(dout, type(dates[0]))
        dsoll1 = dsoll[:16]
        assert dout == dsoll1
        assert xout == xsoll

        # dates = pyjams.datetime
        itest += 1
        print('itest', itest)
        dates = self.ddates  # datetime
        dout, xout = means(dates, self.x)
        assert isinstance(dout, type(dates[0]))
        dout = dout.strftime(self.format)
        assert dout == dsoll
        assert xout == xsoll

        # dates = cf.datetime
        itest += 1
        print('itest', itest)
        dates = [ cf.datetime(*dd.to_tuple())
                  for dd in self.ddates ]
        dout, xout = means(dates, self.x)
        assert isinstance(dout, type(dates[0]))
        dout = dout.strftime(self.format)
        assert dout == dsoll
        assert xout == xsoll

        # dates = datetime.datetime
        itest += 1
        print('itest', itest)
        dates = [ pydatetime.datetime(*dd.to_tuple())
                  for dd in self.ddates ]
        dout, xout = means(dates, self.x)
        assert isinstance(dout, type(dates[0]))
        dout = dout.strftime(self.format)
        assert dout == dsoll
        assert xout == xsoll

        # dates = numeric
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)  # num
        dout, xout = means(dates, self.x, calendar=cal)
        assert isinstance(dout, (type(dates[0]), np.int64))
        dout = num2date(dout, calendar=cal)
        dout = dout.strftime(self.format)
        assert dout == dsoll
        assert xout == xsoll

        # data = ndarray
        itest += 1
        print('itest', itest)
        dates = self.dates
        dout, xout = means(dates, np.vstack([self.x, self.x]).T)
        assert isinstance(dout, type(dates[0]))
        xsoll1 = [xsoll, xsoll]
        assert dout == dsoll
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll1))

        # sum
        itest += 1
        print('itest', itest)
        dates = self.dates  # str
        dout, xout = means(dates, self.x, sum=True)
        assert isinstance(dout, type(dates[0]))
        xsoll1 = xsoll * n
        assert dout == dsoll
        assert xout == xsoll1

        # min
        itest += 1
        print('itest', itest)
        dates = self.dates  # str
        dout, xout = means(dates, self.x, min=True)
        assert isinstance(dout, type(dates[0]))
        xsoll1 = self.x[0]
        assert dout == dsoll
        assert xout == xsoll1

        # max
        itest += 1
        print('itest', itest)
        dates = self.dates  # str
        dout, xout = means(dates, self.x, max=True)
        assert isinstance(dout, type(dates[0]))
        xsoll1 = self.x[-1]
        assert dout == dsoll
        assert xout == xsoll1

        # onlydat
        itest += 1
        print('itest', itest)
        dates = self.dates
        xout = means(dates, self.x, onlydat=True)
        assert xout == xsoll

        # onlydat with dat = ndarray
        itest += 1
        print('itest', itest)
        dates = self.dates
        xout = means(dates, np.vstack([self.x, self.x]).T, onlydat=True)
        xsoll1 = [xsoll, xsoll]
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll1))

        # masked
        itest += 1
        print('itest', itest)
        dates = self.dates
        n2 = self.n // 2
        xout = means(dates,
                     np.ma.array(self.x, mask=(self.x > n2)), onlydat=True)
        xsoll1 = n2//2 * (n2 + 1) / n2
        assert xout == xsoll1

        # different means

        # year
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)
        dout, xout = means(dates, self.x, year=True, calendar=cal)
        assert isinstance(dout[0], (type(dates[0]), np.int64))
        dsoll = [ '{:04d}-06-15 12:00:00'.format(self.year0 + yy)
                  for yy in range(self.nyears) ]
        n = self.ndays * self.nhours * self.nminutes
        xsoll = (n//2 * (n + 1) + np.arange(self.nyears) * n * n) / n
        # ToDo: output type wrong
        # assert isinstance(dout, list)
        assert isinstance(dout, np.ndarray)
        assert isinstance(xout, np.ndarray)
        dout = num2date(dout, calendar=cal)
        dout = [ dd.strftime(self.format) for dd in dout ]
        self.assertEqual(_flatten(dout),
                         _flatten(dsoll))
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll))

        # year with dat = ndarray
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)
        dout, xout = means(dates, np.vstack([self.x, self.x]).T,
                           year=True, calendar=cal)
        assert isinstance(dout[0], (type(dates[0]), np.int64))
        dsoll = [ '{:04d}-06-15 12:00:00'.format(self.year0 + yy)
                  for yy in range(self.nyears) ]
        n = self.ndays * self.nhours * self.nminutes
        xsoll = (n//2 * (n + 1) + np.arange(self.nyears) * n * n) / n
        xsoll1 = np.vstack([xsoll, xsoll]).T
        # ToDo: output type wrong
        # assert isinstance(dout, list)
        assert isinstance(dout, np.ndarray)
        assert isinstance(xout, np.ndarray)
        dout = num2date(dout, calendar=cal)
        dout = [ dd.strftime(self.format) for dd in dout ]
        self.assertEqual(_flatten(dout),
                         _flatten(dsoll))
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll1))

        # year and onlydat
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)
        xout = means(dates, self.x, year=True, calendar=cal, onlydat=True)
        n = self.ndays * self.nhours * self.nminutes
        xsoll = (n//2 * (n + 1) + np.arange(self.nyears) * n * n) / n
        assert isinstance(xout, np.ndarray)
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll))

        # year and onlydat and masked array
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)
        xout = means(dates, np.ma.array(self.x), year=True,
                     calendar=cal, onlydat=True)
        n = self.ndays * self.nhours * self.nminutes
        xsoll = (n//2 * (n + 1) + np.arange(self.nyears) * n * n) / n
        assert isinstance(xout, np.ndarray)
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll))

        # year and onlydat and mask
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)
        n2 = self.n // 2
        xout = means(dates, np.ma.array(self.x, mask=(self.x > n2)),
                     year=True, calendar=cal, onlydat=True)
        assert isinstance(xout, np.ma.MaskedArray)
        n = self.ndays * self.nhours * self.nminutes
        assert np.ma.count_masked(xout) > 0
        assert not np.ma.is_masked(xout[0])
        assert np.ma.is_masked(xout[1])
        xsoll = n//2 * (n + 1) / n
        assert xout[0] == xsoll

        # year and onlydat with dat = ndarray
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)
        xout = means(dates, np.vstack([self.x, self.x]).T,
                     year=True, calendar=cal, onlydat=True)
        n = self.ndays * self.nhours * self.nminutes
        xsoll = (n//2 * (n + 1) + np.arange(self.nyears) * n * n) / n
        xsoll1 = np.vstack([xsoll, xsoll]).T
        assert isinstance(xout, np.ndarray)
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll1))

        # month=True

        # day
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)
        dout, xout = means(dates, self.x, day=True, calendar=cal)
        assert isinstance(dout[0], (type(dates[0]), np.int64))
        dlen = self.nhours * self.nminutes
        dsoll = [ dd[:10] + ' 12:00:00'
                  for dd in self.dates[::dlen] ]
        xsoll = np.mean(np.array(self.x).reshape(-1, dlen), axis=1)
        # assert isinstance(dout, list)
        assert isinstance(dout, np.ndarray)
        assert isinstance(xout, np.ndarray)
        dout = num2date(dout, calendar=cal)
        dout = [ dd.strftime(self.format) for dd in dout ]
        self.assertEqual(_flatten(dout),
                         _flatten(dsoll))
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll))

        # hour
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)
        dout, xout = means(dates, self.x, hour=True, calendar=cal)
        assert isinstance(dout[0], (type(dates[0]), np.int64))
        dlen = self.nminutes
        dsoll = [ dd[:14] + '30:00' for dd in self.dates[::dlen] ]
        xsoll = np.mean(np.array(self.x).reshape(-1, dlen), axis=1)
        # assert isinstance(dout, list)
        assert isinstance(dout, np.ndarray)
        assert isinstance(xout, np.ndarray)
        dout = num2date(dout, calendar=cal)
        dout = [ dd.strftime(self.format) for dd in dout ]
        self.assertEqual(_flatten(dout),
                         _flatten(dsoll))
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll))

        # half_hour
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)
        dout, xout = means(dates, self.x, half_hour=True,
                           calendar=cal)
        assert isinstance(dout[0], (type(dates[0]), np.int64))
        dsoll = self.dates
        xsoll = np.array(self.x, dtype=float)
        # assert isinstance(dout, list)
        assert isinstance(dout, np.ndarray)
        assert isinstance(xout, np.ndarray)
        dout = num2date(dout, calendar=cal)
        dout = [ dd.strftime(self.format) for dd in dout ]
        self.assertEqual(_flatten(dout),
                         _flatten(dsoll))
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll))

        # minutes=True takes very long
        # dsoll = self.dates
        # xsoll = np.array(self.x, dtype=float)

        # meanday
        itest += 1
        print('itest', itest)
        cal = calendars[itest % ncal]
        dates = date2num(self.dates, calendar=cal)
        dout, xout = means(dates, self.x, meanday=True,
                           calendar=cal)
        assert isinstance(dout[0], (type(dates[0]), np.int64))
        dsoll = [ '{:04d}-01-01 {:02d}:30:00'.format(self.year0, dd)
                  for dd in range(24) ]
        dlen = self.nhours * self.nminutes
        xsoll = [ np.mean(self.x[i::dlen]) for i in range(dlen) ]
        xsoll = np.mean(np.array(xsoll).reshape(-1, self.nminutes), axis=1)
        # assert isinstance(dout, list)
        assert isinstance(dout, np.ndarray)
        assert isinstance(xout, (np.ndarray, np.ma.MaskedArray))
        dout = num2date(dout, calendar=cal)
        dout = [ dd.strftime(self.format) for dd in dout ]
        self.assertEqual(_flatten(dout),
                         _flatten(dsoll))
        self.assertEqual(_flatten(xout),
                         _flatten(xsoll))

        # meanmonth=True == seasonalmonth=True

        # seasonalday=True

        # seasonalmeanday=True


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
This is the unittest for division module.

python -m unittest -v tests/test_division.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_division.py

"""
from __future__ import division, absolute_import, print_function
import unittest


class TestDivision(unittest.TestCase):
    """
    Tests for division.py
    """

    def test_division(self):
        import numpy as np
        from pyjams import division

        a = (1., 2., 3.)
        b = 2.
        assert isinstance(division(a, b), tuple)
        self.assertEqual(list(division(a, b)),
                         [0.5, 1.0, 1.5])

        a = [1., 2., 3.]
        b = 2.
        assert isinstance(division(a, b), list)
        self.assertEqual(list(division(a, b)),
                         [0.5, 1.0, 1.5])

        a = np.array([1., 2., 3.])
        b = 2.
        assert isinstance(division(a, b), np.ndarray)
        self.assertEqual(list(division(a, b)),
                         [0.5, 1.0, 1.5])

        a = 1.
        b = (1., 2., 4.)
        assert isinstance(division(a, b), tuple)
        self.assertEqual(list(division(a, b)),
                         [1.0, 0.5, 0.25])

        a = 1.
        b = [1., 2., 4.]
        assert isinstance(division(a, b), list)
        self.assertEqual(list(division(a, b)),
                         [1.0, 0.5, 0.25])

        a = 1.
        b = np.array([1., 2., 4.])
        assert isinstance(division(a, b), np.ndarray)
        self.assertEqual(list(division(a, b)),
                         [1.0, 0.5, 0.25])

        a = [1., 1., 1.]
        b = [2., 1., 0.]
        dd = division(a, b)
        assert isinstance(dd, list)
        self.assertEqual(list(dd[0:2]), [0.5, 1.0])
        assert np.isnan(dd[2])
        self.assertEqual(list(division(a, b, 0.)),
                         [0.5, 1.0, 0.0])
        self.assertEqual(list(division(a, b, otherwise=0.)),
                         [0.5, 1.0, 0.0])
        dd = division(a, b, prec=1.)
        self.assertEqual(dd[0], 0.5)
        assert np.isnan(dd[1])
        assert np.isnan(dd[2])

        a = np.array([1., 1., 1.])
        b = [2., 1., 0.]
        dd = division(a, b)
        assert isinstance(dd, np.ndarray)
        self.assertEqual(list(dd[0:2]), [0.5, 1.0])
        assert np.isnan(dd[2])

        a = np.array([1., 1., 1.])
        b = np.array([2., 1., 0.])
        assert isinstance(dd, np.ndarray)
        self.assertEqual(list(dd[0:2]), [0.5, 1.0])
        assert np.isnan(dd[2])

        mask = [0, 0, 1]
        a = np.ma.array([1., 1., 1.], mask=mask)
        b = np.array([2., 1., 0.])
        dd = division(a, b)
        assert isinstance(dd, np.ma.masked_array)
        self.assertEqual(list(dd), [0.5, 1.0, np.ma.masked])

        a = np.array([1., 1., 1.])
        mask = [0, 0, 1]
        b = np.ma.array([2., 1., 0.], mask=mask)
        dd = division(a, b)
        assert isinstance(dd, np.ma.masked_array)
        self.assertEqual(list(dd), [0.5, 1.0, np.ma.masked])

    def test_div(self):
        import numpy as np
        from pyjams import div

        a = (1., 2., 3.)
        b = 2.
        assert isinstance(div(a, b), tuple)
        self.assertEqual(list(div(a, b)),
                         [0.5, 1.0, 1.5])

        a = [1., 2., 3.]
        b = 2.
        assert isinstance(div(a, b), list)
        self.assertEqual(list(div(a, b)),
                         [0.5, 1.0, 1.5])

        a = np.array([1., 2., 3.])
        b = 2.
        assert isinstance(div(a, b), np.ndarray)
        self.assertEqual(list(div(a, b)),
                         [0.5, 1.0, 1.5])

        a = 1.
        b = (1., 2., 4.)
        assert isinstance(div(a, b), tuple)
        self.assertEqual(list(div(a, b)),
                         [1.0, 0.5, 0.25])

        a = 1.
        b = [1., 2., 4.]
        assert isinstance(div(a, b), list)
        self.assertEqual(list(div(a, b)),
                         [1.0, 0.5, 0.25])

        a = 1.
        b = np.array([1., 2., 4.])
        assert isinstance(div(a, b), np.ndarray)
        self.assertEqual(list(div(a, b)),
                         [1.0, 0.5, 0.25])

        a = [1., 1., 1.]
        b = [2., 1., 0.]
        dd = div(a, b)
        assert isinstance(dd, list)
        self.assertEqual(list(dd[0:2]), [0.5, 1.0])
        assert np.isnan(dd[2])
        self.assertEqual(list(div(a, b, 0.)),
                         [0.5, 1.0, 0.0])
        self.assertEqual(list(div(a, b, otherwise=0.)),
                         [0.5, 1.0, 0.0])
        dd = div(a, b, prec=1.)
        self.assertEqual(dd[0], 0.5)
        assert np.isnan(dd[1])
        assert np.isnan(dd[2])

        a = np.array([1., 1., 1.])
        b = [2., 1., 0.]
        dd = div(a, b)
        assert isinstance(dd, np.ndarray)
        self.assertEqual(list(dd[0:2]), [0.5, 1.0])
        assert np.isnan(dd[2])

        a = np.array([1., 1., 1.])
        b = np.array([2., 1., 0.])
        assert isinstance(dd, np.ndarray)
        self.assertEqual(list(dd[0:2]), [0.5, 1.0])
        assert np.isnan(dd[2])

        mask = [0, 0, 1]
        a = np.ma.array([1., 1., 1.], mask=mask)
        b = np.array([2., 1., 0.])
        dd = div(a, b)
        assert isinstance(dd, np.ma.masked_array)
        self.assertEqual(list(dd), [0.5, 1.0, np.ma.masked])

        a = np.array([1., 1., 1.])
        mask = [0, 0, 1]
        b = np.ma.array([2., 1., 0.], mask=mask)
        dd = div(a, b)
        assert isinstance(dd, np.ma.masked_array)
        self.assertEqual(list(dd), [0.5, 1.0, np.ma.masked])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
Write text on a plot

This module was written by Matthias Cuntz while at Department of Computational
Hydrosystems, Helmholtz Centre for Environmental Research - UFZ, Leipzig,
Germany, and continued while at Institut National de Recherche pour
l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2014-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   text2plot
   abc2plot
   signature2plot

History
    * Written Nov 2021 by Matthias Cuntz (mc (at) macu (dot) de)
      combining abc2plot and signature2plot
    * Written abc2plot, May 2012, Matthias Cuntz
    * Added parenthesis option to abc2plot, Feb 2013, Arndt Piayda
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Added opening and closing parentheses, brackets, braces,
      Feb 2013, Matthias Cuntz
    * Options usetex and mathrm, Feb 2013, Matthias Cuntz
    * Options mathbf, large and making medium the default,
      Feb 2013, Matthias Cuntz
    * Added string option, Nov 2013, Matthias Cuntz
    * Corrected bug in medium as default, Nov 2013, Matthias Cuntz
    * Make usetex work with fontsize keyword of axis.text()
      of matplotllib v1.1.0, Nov 2013, Matthias Cuntz
    * Written signature2plot, Jan 2014, Matthias Cuntz
    * Assert that small or large is set if medium is not None (abc2plot),
      Feb 2014, Matthias Cuntz
    * Replace horizontalalignment='left' and verticalalignment='bottom'
      by kwargs mechanism (abc2plot), May 2014, Matthias Cuntz
    * Added option italic (abc2plot), May 2014, Matthias Cuntz
    * Added options xlarge, xxlarge, xsmall, xxsmall (abc2plot),
      Oct 2015, Matthias Cuntz
    * Make numpy docstring format (abc2plot), Nov 2020, Matthias Cuntz
    * Ported into pyjams, Nov 2021, Matthias Cuntz
    * dx, dy, and name mandatory parameters (signature2plot),
      Nov 2020, Matthias Cuntz
    * Change option name parenthesis -> parentheses, Nov 2021, Matthias Cuntz
    * Written text2plot, Nov 2021, Matthias Cuntz
    * Use text2plot for signature2plot, Nov 2021, Matthias Cuntz
    * Use text2plot for abc2plot, Nov 2021, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""
import time as ptime
# from .str2tex import str2tex
# from .romanliterals import int2roman
from pyjams.str2tex import str2tex
from pyjams.romanliterals import int2roman


__all__ = ['text2plot', 'abc2plot', 'signature2plot']


def text2plot(handle, dx, dy, itext,
              small=False, medium=False, large=False,
              xsmall=False, xxsmall=False, xlarge=False, xxlarge=False,
              bold=False, italic=False,
              usetex=False, mathrm=False,
              **kwargs):  # pragma: no cover
    """
    Write text on plot

    Parameters
    ----------
    handle : :class:`matplotlib.axes` subclass
        Matplotlib axes handle
    dx : float
        % of xlim from min(xlim)
    dy : float
        % of ylim from min(ylim)
    itext : str
        String to write on plot
    small : bool, optional
        fontsize='small' if True (default: False)
    medium : bool, optional
        fontsize='medium' if True (default: False).
        Medium is taken if no other fontsize is chosen.
    large : bool, optional
        fontsize='large' if True (default: False)
    xlarge : bool, optional
        fontsize='x-large' if True (default: False)
    xsmall : bool, optional
        fontsize='x-small' if True (default: False)
    xxlarge : bool, optional
        fontsize='xx-large' if True (default: False)
    xxsmall : bool, optional
        fontsize='xx-small' if True (default: False)
    bold : bool, optional
        fontweight='bold' if True, else fontsize='normal' (default)
    italic : bool, optional
        fontstyle='italic' if True, else fontstyle='normal' (default)
    usetex : bool, optional
        Embed into LaTeX math environment if True,
        else no LaTeX math mode (default)
    mathrm : bool, optional
        If True, put text into appropriate LaTeX mathrm/mathit/mathbf
        environment if *usetex==True* and *italic==True* or *bold==True*.
        If False, use standard math font if *usetex==True* (default).
    string : bool, optional
        Treat *iplot* as literal string and not as integer if True (default:
        False). *integer*, *roman* and *lower* are disabled then.
    **kwargs : dict, optional
        All additional parameters are passed passed to
        :meth:`matplotlib.axes.Axes.text()`

    Returns
    -------
    String on plot

    Examples
    --------
    .. code-block:: python

       text2plot(ax, 0.7, 0.6, r'CO$_2$', large=True,
                 usetex=usetex, mathrm=False)

    """
    import matplotlib.pyplot as plt

    # Check input
    ifont = small + medium + large + xsmall + xxsmall + xlarge + xxlarge
    assert ifont <= 1, ('only one of small, medium, large, xsmall, xxsmall'
                        ' xlarge, or xxlarge font size can be chosen.')
    if ifont == 0:
        medium = True
    if usetex and mathrm:
        assert (bold + italic) <= 1, ('if usetex and mathrm: then bold and'
                                      ' italic are mutually exclusive.')

    # Size
    if small:
        fs = 'small'
    elif medium:
        fs = 'medium'
    elif large:
        fs = 'large'
    elif xsmall:
        fs = 'x-small'
    elif xxsmall:
        fs = 'xx-small'
    elif xlarge:
        fs = 'x-large'
    elif xxlarge:
        fs = 'xx-large'
    else:  # just for security
        fs = 'medium'

    # Weight
    if bold:
        fw = 'bold'
    else:
        fw = 'normal'

    # Style
    if italic:
        fst = 'italic'
    else:
        fst = 'normal'

    # usetex
    istr = str2tex(itext, usetex=usetex)

    if usetex:
        if mathrm:
            if bold:
                istr = istr.replace('mathrm', 'mathbf')
            elif italic:
                istr = istr.replace('mathrm', 'mathit')

    # x/y position
    xmin, xmax = handle.get_xlim()
    ymin, ymax = handle.get_ylim()
    idx = xmin + dx * (xmax - xmin)
    if 'transform' in kwargs:
        if kwargs['transform'] is handle.transAxes:
            idx = dx
    idy = ymin + dy * (ymax - ymin)
    if 'transform' in kwargs:
        if kwargs['transform'] is handle.transAxes:
            idy = dy

    handle.text(idx, idy, istr,
                fontsize=fs, fontweight=fw, fontstyle=fst,
                **kwargs)


def abc2plot(handle, dx, dy, iplot,
             integer=False, roman=False, lower=False,
             parentheses=None, brackets=None, braces=None,
             bold=False, italic=False,
             usetex=False, mathrm=False, string=False,
             **kwargs):  # pragma: no cover
    """
    Write a, B, iii, IV, e), etc. on plots

    Parameters
    ----------
    handle : :class:`matplotlib.axes` subclass
       Matplotlib axes handle
    dx : float
        % of xlim from min(xlim)
    dy : float
        % of ylim from min(ylim)
    iplot : int or str
        Number of plot starting with 1, or string if 'string==True'
    integer : bool, optional
        Use integers instead of a, b, c if True (default: False)
    roman : bool, optional
        Use Roman literals instead of a, b, c if True (default: False)
    lower : bool, optional
        Use lowercase letters for a, b, c if True,
        else use uppercase letters (default)
    parentheses : str or None, optional
        Parentheses before or after the letter/number. Possible values are
        'open', 'close', 'both', 'None', and *None*.
        'open' puts opening parentheses in front of the letter/number.
        'close' puts closing parentheses after the letter/number.
        'both' puts opening and closing parentheses around the letter/number.
        'None' and *None* do not put any parentheses before or after the
        letter/number (default).
    brackets : str or None, optional
        Brackets before or after the letter/number. Possible values are
        'open', 'close', 'both', 'None', and *None*.
        'open' puts opening brackets in front of the letter/number.
        'close' puts closing brackets after the letter/number.
        'both' puts opening and closing brackets around the letter/number.
        'None' and *None* do not put any brackets before or after the
        letter/number (default).
    braces : str or None, optional
        Braces before or after the letter/number. Possible values are
        'open', 'close', 'both', 'None', and *None*.
        'open' puts opening braces in front of the letter/number.
        'close' puts closing braces after the letter/number.
        'both' puts opening and closing braces around the letter/number.
        'None' and *None* do not put any braces before or after the
        letter/number (default).
    bold : bool, optional
        fontweight='bold' if True, else fontsize='normal' (default)
    italic : bool, optional
        fontstyle='italic' if True, else fontstyle='normal' (default)
    usetex : bool, optional
        Embed into LaTeX math environment if True,
        else no LaTeX math mode (default)
    mathrm : bool, optional
        If True, put text into appropriate LaTeX mathrm/mathit/mathbf
        environment if *usetex==True* and *italic==True* or *bold==True*.
        If False, use standard math font if *usetex==True* (default).
    string : bool, optional
        Treat *iplot* as literal string and not as integer if True (default:
        False). *integer*, *roman* and *lower* are disabled then.
    **kwargs : dict, optional
        All additional parameters are passed passed to :func:`text2plot()`

    Returns
    -------
    Letter/number or string on plot

    Notes
    -----
    If output is a letter then iplot > 26 gives unexpected results.

    Examples
    --------
    .. code-block:: python

       abc2plot(ax, 0.7, 0.6, 2, large=True, parentheses='both',
                usetex=usetex, mathrm=False)

    """
    import matplotlib.pyplot as plt

    # Check input
    assert (roman + integer) < 2, ('either Roman literals or integers can'
                                   ' be chosen.')
    iparentheses = False
    if parentheses is not None:
        if parentheses != 'None':
            iparentheses = True
    ibrackets = False
    if brackets is not None:
        if brackets != 'None':
            ibrackets = True
    ibraces = False
    if braces is not None:
        if braces != 'None':
            ibraces = True
    ibrack = iparentheses + ibrackets + ibraces
    assert ibrack <= 1, ('only one of parentheses, brackets, or braces'
                         ' can be chosen.')

    # Number or letter
    if string:
        t = str(iplot)
    else:
        if roman:
            t = int2roman(iplot, lower=lower)
        elif integer:
            t = str(iplot)
        else:
            if lower:
                t = chr(96+iplot)
            else:
                t = chr(64+iplot)

    # parentheses
    if iparentheses:
        if parentheses.lower() == 'open':
            t = '(' + t
        elif parentheses.lower() == 'close':
            t = t + ')'
        elif parentheses.lower() == 'both':
            t = '(' + t + ')'
        elif parentheses.lower() == 'none':
            pass
        else:
            raise ValueError("parentheses must be either 'open', 'close',"
                             " 'both', or 'none'.")

    if ibrackets:
        if brackets.lower() == 'open':
            t = '[' + t
        elif brackets.lower() == 'close':
            t = t + ']'
        elif brackets.lower() == 'both':
            t = '[' + t + ']'
        elif brackets.lower() == 'none':
            pass
        else:
            raise ValueError("brackets must be either 'open', 'close',"
                             " 'both', or 'none'.")

    if ibraces:
        if braces.lower() == 'open':
            if usetex or (plt.get_backend() == 'pdf'):
                t = r'\{' + t
            else:
                t = '{' + t
        elif braces.lower() == 'close':
            if usetex or (plt.get_backend() == 'pdf'):
                t = t + r'\}'
            else:
                t = t + '}'
        elif braces.lower() == 'both':
            if usetex or (plt.get_backend() == 'pdf'):
                t = r'\{' + t + r'\}'
            else:
                t = '{' + t + '}'
        elif braces.lower() == 'none':
            pass
        else:
            raise ValueError("braces must be either 'open', 'close',"
                             " 'both', or 'none'.")

    if usetex:
        if mathrm:
            if bold:
                t = r'\mathbf{' + t + '}'
            elif italic:
                t = r'\mathit{' + t + '}'
            else:
                t = r'\mathrm{' + t + '}'
        t = r'$' + t + r'$'

    text2plot(handle, dx, dy, t,
              bold=bold, italic=italic,
              usetex=False, mathrm=False,
              **kwargs)


def signature2plot(handle, dx, dy, itext,
                   **kwargs):  # pragma: no cover
    """
    Write a copyright notice on a plot

    The notice is: '(C) YYYY *itext*', where YYYY is the current 4-digit year.

    Parameters
    ----------
    handle : :class:`matplotlib.axes` subclass
       Matplotlib axes handle
    dx : float
        % of xlim from min(xlim)
    dy : float
        % of ylim from min(ylim)
    itext : str
        Text after '(C) YYYY', where YYYY is the current 4-digit year
    **kwargs : dict, optional
        All additional parameters are passed passed to :func:`text2plot()`

    Returns
    -------
    '(C) YYYY itext' or '(C) YYYY' on plot, where YYYY is the 4-digit year

    Notes
    -----
    Signature will be right-aligned horizontally if horizontalalignment or
    ha are not given in kwargs.

    Examples
    --------
    .. code-block:: python

       signature2plot(ax, 0., 0.05, 'M Cuntz, INRAE',
                      small=True, italic=True,
                      usetex=usetex, mathrm=True)

    """
    year = str(ptime.localtime().tm_year)
    if itext.strip():
        otext = r'$\copyright$ ' + year + ' ' + itext.strip()
    else:
        otext = r'$\copyright$ ' + year

    if ('horizontalalignment' not in kwargs) and ('ha' not in kwargs):
        kwargs['horizontalalignment'] = 'right'

    text2plot(handle, dx, dy, otext, **kwargs)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # outtype = ''
    # # outtype = 'pdf'
    # pdffile = 'text2plot.pdf'
    # usetex  = True
    # textsize = 12

    # import matplotlib as mpl
    # if (outtype == 'pdf'):
    #     mpl.use('PDF')  # set directly after import matplotlib
    #     import matplotlib.pyplot as plt
    #     from matplotlib.backends.backend_pdf import PdfPages
    #     # Customize: http://matplotlib.sourceforge.net/users/customizing.html
    #     mpl.rc('ps', papersize='a4', usedistiller='xpdf')  # ps2pdf
    #     mpl.rc('figure', figsize=(8.27, 11.69))  # a4 portrait
    #     if usetex:
    #         mpl.rc('text', usetex=True)
    # else:
    #     import matplotlib.pyplot as plt
    #     mpl.rc('figure', figsize=(4./5.*8.27, 4./5.*11.69))  # a4 portrait
    # mpl.rc('path', simplify=False)  # do not remove

    # if (outtype == 'pdf'):
    #     print('Plot PDF ', pdffile)
    #     pdf_pages = PdfPages(pdffile)
    # else:
    #     print('Plot X')
    # figsize = mpl.rcParams['figure.figsize']

    # # --------------------------------------------------------------
    # # text2plot
    # fig = plt.figure()
    # sub = fig.add_axes([0.05, 0.05, 0.4, 0.4])
    # m = plt.plot(range(100), 'k:')
    # text2plot(sub, 0.0, 0.0, r'CO$_2$')
    # text2plot(sub, 0.1, 0.1, 'CO2', usetex=usetex)
    # text2plot(sub, 0.2, 0.2, r'$\frac{CO_2}{O_2}$', xxsmall=True,
    #           usetex=usetex, mathrm=False)
    # text2plot(sub, 0.3, 0.3, r'Units (m$^2$ s$^{-1}$)', xsmall=True,
    #           usetex=usetex, mathrm=False)
    # text2plot(sub, 0.4, 0.4, r'Two \n lines', medium=True,
    #           usetex=False, mathrm=False)
    # text2plot(sub, 0.5, 0.5, r'Units \newline (m$^2$ s$^{-1}$)', large=True,
    #           usetex=usetex, mathrm=False)
    # text2plot(sub, 0.6, 0.6, r'CO$_2$', xlarge=True,
    #           usetex=usetex, mathrm=False)
    # text2plot(sub, 0.7, 0.7, r'CO$_2$', xxlarge=True,
    #           usetex=usetex, mathrm=False)
    # text2plot(sub, 0.8, 0.8, r'CO$_2$', medium=True, bold=True,
    #           usetex=usetex, mathrm=True)

    # # --------------------------------------------------------------
    # # abc2plot
    # fig = plt.figure()
    # sub = fig.add_axes([0.05, 0.05, 0.4, 0.4])
    # m = plt.plot(range(100), 'k:')
    # abc2plot(sub, 0, 0, 2)
    # abc2plot(sub, 0.1, 0.1, 2, parentheses='close')
    # abc2plot(sub, 0.2, 0.2, 2, lower=True, parentheses='open')
    # abc2plot(sub, 0.3, 0.3, 2, roman=True, parentheses='both')
    # abc2plot(sub, 0.4, 0.4, 2, roman=True, lower=True, parentheses='none')
    # abc2plot(sub, 0.5, 0.5, 2, integer=True, parentheses='both',
    #          usetex=usetex)
    # abc2plot(sub, 0.5, 0.6, 2, small=True, medium=False,
    #          large=False, parentheses='both', usetex=usetex,
    #          mathrm=False)
    # abc2plot(sub, 0.6, 0.6, 2, small=False, medium=True,
    #          large=False, parentheses='both', usetex=usetex,
    #          mathrm=False)
    # abc2plot(sub, 0.7, 0.6, 2, small=False, medium=False,
    #          large=True, parentheses='both', usetex=usetex,
    #          mathrm=False)
    # abc2plot(sub, 0.7, 0.7, 2, medium=True, brackets='both',
    #          usetex=usetex, mathrm=True)
    # abc2plot(sub, 0.8, 0.8, 2, medium=True, bold=True,
    #          braces='both', usetex=usetex, mathrm=True)

    # sub = fig.add_axes([0.5, 0.5, 0.4, 0.4])
    # m = plt.plot(range(100), 'k:')
    # abc2plot(sub, 0.1, 0.1, 2, brackets='close')
    # abc2plot(sub, 0.2, 0.2, 2, lower=True, brackets='open')
    # abc2plot(sub, 0.3, 0.3, 2, roman=True, brackets='both',
    #          usetex=usetex, mathrm=False)
    # abc2plot(sub, 0.4, 0.4, 2, roman=True, lower=True,
    #          brackets='none')
    # abc2plot(sub, 0.5, 0.5, 2, integer=True, braces='close')
    # abc2plot(sub, 0.6, 0.6, 2, small=True, braces='open',
    #          usetex=usetex, mathrm=True)
    # abc2plot(sub, 0.7, 0.7, 2, medium=True, braces='both',
    #          horizontalalignment='left', verticalalignment='bottom')
    # abc2plot(sub, 0.8, 0.8, 2, medium=True, bold=True, braces='none',
    #          horizontalalignment='right', verticalalignment='top')

    # # --------------------------------------------------------------
    # # signature2plot
    # import numpy as np
    # fig = plt.figure()
    # sub = fig.add_axes([0.05, 0.05, 0.4, 0.4])
    # mulx = 1.
    # muly = 1.
    # m = plt.plot(mulx*np.arange(100)/99., muly*np.arange(100)/99., 'k:')
    # signature2plot(sub, mulx*0.5, muly*0.5, 'Test1', usetex=usetex)
    # signature2plot(sub, mulx*0.6, muly*0.6, 'MC1', small=True,
    #                usetex=usetex)  # -
    # signature2plot(sub, mulx*0.7, muly*0.7, 'Test2', large=True,
    #                usetex=usetex, italic=True)
    # signature2plot(sub, mulx*0.8, muly*0.8, 'MC2', bold=True, usetex=usetex,
    #                ha='left', mathrm=True)
    # signature2plot(sub, mulx*0.8, muly*0.8, 'MC2', bold=True)
    # signature2plot(sub, mulx*0.9, muly*0.9, 'Test3', large=True,
    #                usetex=usetex, italic=True, mathrm=True)
    # signature2plot(sub, mulx*1.0, muly*1.0, 'MC3', usetex=usetex, bold=True,
    #                horizontalalignment='left')
    # signature2plot(sub, 0.9, 0.1, 'MM', transform=sub.transAxes,
    #                usetex=usetex, bold=True, horizontalalignment='center')

    # if (outtype == 'pdf'):
    #     pdf_pages.savefig(fig)
    #     plt.close()

    # if (outtype == 'pdf'):
    #     pdf_pages.close()
    # else:
    #     plt.show()



#!/usr/bin/env python
"""
Get variables from or print information of a netcdf file

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2009-2022 Matthias Cuntz, Stephan Thober, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   infonetcdf
   ncinfo
   readnetcdf
   ncread

History
    * Written Jul 2009 by
      Matthias Cuntz (mc (at) macu (dot) de)
    * Removed quiet keyword, Jun 2012, Matthias Cuntz
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Wrapper functions netcdfread, ncread, readnc, Oct 2013, Matthias Cuntz
    * overwrite keyword, Apr 2014, Stephan Thober
    * dims keyword, May 2014, Stephan Thober
    * attributes keyword, Jun 2016, Stephan Thober
    * Restrict overwrite to files with only one variables,
      Aug 2016, Stephan Thober
    * Do not count dimension variables in variable count for overwrite,
      Oct 2016, Matthias Cuntz
    * Remove wrappers netcdfread and readnc, Mar 2022, Matthias Cuntz
    * Remove reform keyword, Mar 2022, Matthias Cuntz
    * Put all info abilities into separate routine infonetcdf
      with wrapper function ncinfo, Mar 2022, Matthias Cuntz

"""


__all__ = ['infonetcdf', 'ncinfo',
           'readnetcdf', 'ncread']


def infonetcdf(ncfile,
               var='', code=-1, dims=False, shape=False, attributes=False,
               variables=False, codes=False,
               long_names=False, units=False,
               sort=False):
    """
    Get information on variables in a netcdf file

    Parameters
    ----------
    ncfile : str
        netCDF file name
    var : str, optional
        Variable name, only relevant if *dims* or *attributes* are True
        *var* takes precedence over *code*.
    code : int, optional
        Variable code such as in files coming from GRIB,
        only relevant if *dims* or *attributes* are True.
        *var* takes precedence over *code*.
    dims : bool, optional
        Get tuple of dimension names for the variable with name *var*
        or number *code*.
    shape : bool, optional
        Get shape of the variable with name *var* or number *code*.
    attributes : bool, optional
        Get dictionary of all attributes of variable with name *var* or number
        *code*, or all file attributes of *ncfile* if *var* and *code* are not
        given
    variables : bool, optional
        Get list of variables in *ncfile*
    codes : bool, optional
        Get list of variable attributes *code*
        Missing codes will be filled with -1.
    long_names : bool, optional
        Get list of variable attributes *long_name*
        Missing long_names will be filled with ''.
    units : bool, optional
        Get list of variable attributes *units*.
        Missing units will be filled with ''.
    sort : bool, optional
        Sort output of *variables*, *codes*, *units*, and *long_names*
        with variable name as the sort key

    Returns
    -------
    tuple of variable dimension names,
    tuple of variable dimensions,
    list of variable names, codes, units, long_names, or
    dictionary of attributes


    Examples
    --------

    Get variable names

    >>> ncfile = 'test_readnetcdf.nc'
    >>> print([ str(i) for i in infonetcdf(ncfile, variables=True) ])
    ['x', 'y', 'is1', 'is2']
    >>> print([ str(i)
    ...        for i in infonetcdf(ncfile, variables=True, sort=True) ])
    ['is1', 'is2', 'x', 'y']

    Get codes

    >>> print(infonetcdf(ncfile, codes=True))
    [-1, -1, 128, 129]
    >>> print(infonetcdf(ncfile, codes=True, sort=True))
    [128, 129, -1, -1]

    Get special attributes units and long_names

    >>> print([ str(i) for i in infonetcdf(ncfile, units=True) ])
    ['xx', 'yy', 'arbitrary', 'arbitrary']
    >>> print([ str(i) for i in infonetcdf(ncfile, units=True, sort=True) ])
    ['arbitrary', 'arbitrary', 'xx', 'yy']
    >>> print([ str(i) for i in infonetcdf(ncfile, long_names=True) ])
    ['x-axis', 'y-axis', 'all ones', 'all twos']
    >>> print([ str(i)
    ...        for i in infonetcdf(ncfile, long_names=True, sort=True) ])
    ['all ones', 'all twos', 'x-axis', 'y-axis']

    Get dims

    >>> print([ str(i) for i in infonetcdf(ncfile, var='is1', dims=True) ])
    ['y', 'x']

    Get shape

    >>> print(infonetcdf(ncfile, var='is1', shape=True))
    (2, 4)

    Get attributes

    >>> t1 = infonetcdf(ncfile, var='is1', attributes=True)
    >>> print([ str(i) for i in sorted(t1) ])
    ['code', 'long_name', 'units']

    """
    import netCDF4 as nc

    f = nc.Dataset(ncfile, 'r')

    # Variables
    vvars = list(f.variables.keys())
    nvars = len(vvars)

    # Sort and get sort indices
    if sort:
        # index
        ivars = sorted(range(nvars), key=vvars.__getitem__)
        # sorted variables
        svars = [ vvars[i] for i in ivars ]
    else:
        svars = vvars
        ivars = list(range(nvars))

    # variables
    if variables:
        f.close()
        return svars

    # code
    if codes:
        cods = [-1]*nvars
        for i, v in enumerate(svars):
            attr = f.variables[v].ncattrs()
            if 'code' in attr:
                cods[i] = getattr(f.variables[v], 'code')
        f.close()
        return cods

    # long_name
    if long_names:
        lnames = ['']*nvars
        for i, v in enumerate(svars):
            attr = f.variables[v].ncattrs()
            if 'long_name' in attr:
                lnames[i] = getattr(f.variables[v], 'long_name')
        f.close()
        return lnames

    # units
    if units:
        unts = ['']*nvars
        for i, v in enumerate(svars):
            attr = f.variables[v].ncattrs()
            if 'units' in attr:
                unts[i] = getattr(f.variables[v], 'units')
        f.close()
        return unts

    # dims
    if dims:
        if (not var) and (code == -1):
            raise ValueError('var or code has to be given for'
                             ' inquiring dimensions.')
        if var:
            if var not in svars:
                f.close()
                raise ValueError(f'Variable {var} not in file {ncfile}.')
            dimensions = f.variables[var].dimensions
            f.close()
            return dimensions
        if code > -1:
            cods = [-1]*nvars
            for i, v in enumerate(svars):
                attr = f.variables[v].ncattrs()
                if 'code' in attr:
                    cods[i] = getattr(f.variables[v], 'code')
            if code not in cods:
                f.close()
                raise ValueError(f'Code {code} not in file {ncfile}.')
            var = svars[cods.index(code)]
            dimensions = f.variables[var].dimensions
            f.close()
            return dimensions

    # shape
    if shape:
        if (not var) and (code == -1):
            raise ValueError('var or code has to be given for'
                             ' inquiring its shape.')
        if var:
            if var not in svars:
                f.close()
                raise ValueError(f'Variable {var} not in file {ncfile}.')
            shape = f.variables[var].shape
            f.close()
            return shape
        if code > -1:
            cods = [-1]*nvars
            for i, v in enumerate(svars):
                attr = f.variables[v].ncattrs()
                if 'code' in attr:
                    cods[i] = getattr(f.variables[v], 'code')
            if code not in cods:
                f.close()
                raise ValueError(f'Code {code} not in file {ncfile}.')
            var = svars[cods.index(code)]
            shape = f.variables[var].shape
            f.close()
            return shape

    # attributes
    if attributes:
        # file attributes
        if (not var) and (code == -1):
            attrs = dict()
            attr = f.ncattrs()
            for a in attr:
                attrs[a] = getattr(f, a)
            f.close()
            return attrs

        # variables attributes
        if var:
            if var not in svars:
                f.close()
                raise ValueError(f'Variable {var} not in file {ncfile}.')
            attrs = dict()
            attr = f.variables[var].ncattrs()
            for a in attr:
                attrs[a] = getattr(f.variables[var], a)
            f.close()
            return attrs

        if code > -1:
            cods = [-1]*nvars
            for i, v in enumerate(svars):
                attr = f.variables[v].ncattrs()
                if 'code' in attr:
                    cods[i] = getattr(f.variables[v], 'code')
            if code not in cods:
                f.close()
                raise ValueError(f'Code {code} not in file {ncfile}.')
            var = svars[cods.index(code)]
            attrs = dict()
            attr = f.variables[var].ncattrs()
            for a in attr:
                attrs[a] = getattr(f.variables[var], a)
            f.close()
            return attrs

    # no keyword
    f.close()
    return


def ncinfo(*args, **kwargs):
    """
    Wrapper for :func:`infonetcdf`
    """
    return infonetcdf(*args, **kwargs)


def readnetcdf(ncfile, var='', code=-1, squeeze=False,
               pointer=False, overwrite=False):
    """
    Gets variables of a netcdf file

    Parameters
    ----------
    ncfile : str
        netCDF file name
    variables : bool, optional
        Get list of variables in *ncfile*
    codes : bool, optional
        Get list of variable attributes *code*
        Missing codes will be filled with -1.
    squeeze : bool, optional
        Squeeze output array, i.e. remove dimensions of size 1.
    pointer : bool, optional
        Return pointers to the open file and to the variable if True,
        i.e. without actually reading the variable.
        The file will be open in read-only 'r' mode.
        *overwrite* precedes over *pointer*.
    overwrite : bool, optional
        Return pointers to the open file and to the variable if True,
        where the file is opened in append 'a' mode allowing to modify
        the variable. ``readnetcdf`` allows *overwrite* only if the file
        contains a single variable (without the dimension variables).
        *overwrite* precedes over *pointer*.

    Returns
    -------
    numpy array of the variable with name *var* or number *code*, or
    (file pointer, variable pointer)


    Examples
    --------

    Read variable or code

    >>> ncfile = 'test_readnetcdf.nc'
    >>> print(readnetcdf(ncfile, var='is1'))
    [[1. 1. 1. 1.]
     [1. 1. 1. 1.]]
    >>> print(readnetcdf(ncfile, code=129))
    [[2. 2. 2. 2.]
     [2. 2. 2. 2.]]

    Just get file handle so that read is done later at indexing
    useful for example to inquire remote netcdf files first

    >>> fh, var = readnetcdf(ncfile, var='is1', pointer=True)
    >>> print(var.shape)
    (2, 4)
    >>> print(var[:])
    [[1. 1. 1. 1.]
     [1. 1. 1. 1.]]
    >>> fh.close()

    Change a variable in a file

    >>> ncfile = 'test_readnetcdf1.nc'
    >>> print(readnetcdf(ncfile, var='is1'))
    [[1. 1. 1. 1.]
     [1. 1. 1. 1.]]
    >>> fh, var = readnetcdf(ncfile, var='is1', overwrite=True)
    >>> var[:] *= 2.
    >>> fh.close()
    >>> print(readnetcdf(ncfile, var='is1'))
    [[2. 2. 2. 2.]
     [2. 2. 2. 2.]]
    >>> fh, var = readnetcdf(ncfile, var='is1', overwrite=True)
    >>> var[:] *= 0.5
    >>> fh.close()

    """
    import netCDF4 as nc

    if (not var) and (code == -1):
        raise ValueError('var or code has to be given.')

    vvars = infonetcdf(ncfile, variables=True)
    if var:
        if var not in vvars:
            raise ValueError(f'Variable {var} not in file {ncfile}.')
    else:
        cods = infonetcdf(ncfile, codes=True)
        if code not in cods:
            raise ValueError(f'Code {code} not in file {ncfile}.')
        var = vvars[cods.index(code)]

    fmode = 'r'
    if overwrite:
        fmode = 'a'
        # check that only one variable
        dims = infonetcdf(ncfile, var=var, dims=True)
        vvars1 = [ v for v in vvars if v not in dims ]
        if len(vvars1) > 1:
            raise ValueError('overwrite is only allowed on files with a single'
                             ' variable (without dimension variables).')

    f = nc.Dataset(ncfile, fmode)
    v = f.variables[var]

    if pointer or overwrite:
        return f, v

    v = v[:]
    if squeeze:
        v = v.squeeze()

    f.close()
    return v


def ncread(*args, **kwargs):
    """
    Wrapper for :func:`readnetcdf`
    """
    return readnetcdf(*args, **kwargs)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



"""
netCDF4 functions to copy a netcdf file while doing some transformations on
variables and dimensions.

An elaborate example of copying a netcdf file could remove one variable
(removevar), rename another variable (renamevar), replace name and values of a
third variable (replacevar), setting some attributes of old and new variables
(replaceatt), and leave the output file open for further modification
(noclose):

.. code-block:: python

   import numpy as np
   import pyjams.ncio as ncio

   newarr = np.full((720, 360), 273.15)
   fo = ncio.copy_file('infile.nc', 'infile.nc',
                       removevar=['var0'],
                       renamevar={'var1': 'varnew1'},
                       replacevar={'var2': {'varnew2': newvar}},
                       replaceatt={'varnew1': {'long_name': 'renamed var1'},
                                   'varnew2': {'long_name': 'new variable for var2',
                                               'units': 'arbitrary'}},
                       noclose=True)
   # change var3 afterwards
   ovar = fo.variables['var3']
   ovar[:] = ovar[:] * 3.
   fo.close()


Using the individual routines for manipulating dimensions and variables could
copy all variables in a file while making latitudinal means (assume longitudes
in the last dimension), and add a new variable:

.. code-block:: python

   import netCDF4 as nc
   import pyjams.ncio as ncio

   ifile = 'input.nc'  # name of input file
   vtime = 'time'      # name of time variable
   vlon  = 'lon'       # name of longitude variable

   # open files
   ofile = ncio.set_output_filename(ifile, '-latmean.nc')
   fi = nc.Dataset(ifile, 'r')
   if 'file_format' in dir(fi):
       fo = nc.Dataset(ofile, 'w', format=fi.file_format)
   else:
       fo = nc.Dataset(ofile, 'w', format='NETCDF4')
   ntime = fi.dimensions[vtime].size

   # meta data
   ncio.copy_global_attributes(fi, fo, add={'history': 'latitudinal mean'})

   # copy dimensions
   ncio.copy_dimensions(fi, fo, removedim=[vlon])

   # create variables
   # this could be one command (time=None or keyword time left out)
   # but I like to have the non-time dependent variables at the beginning
   # of the netcdf file
   # create static variables (independent of time)
   ncio.create_variables(fi, fo, time=False, timedim=vtime, fill=True,
                         removedim=[vlon])
   # create dynamic variables (time dependent)
   ncio.create_variables(fi, fo, time=True, timedim=vtime, fill=True,
                         removedim=[vlon])

   # create new variable
   dims = list(fi.variables['var1'].dimensions)
   dims = dims[:-1]
   odict = {'name': 'var2',
            'dtype': fi.variables['var1'].dtype,
            'standard_name': 'var2',
            'long_name': '2nd variable',
            'dimensions': dims}
   var2 = ncio.create_new_variable(odict, fo, izip=True, fill=True)

   # copy static variables (not time-dependent) making latitudinal means
   for ivar in fi.variables.values():
       if vtime not in ivar.dimensions:
           ovar = fo.variables[ivar.name]
           if vlon in ivar.dimensions:
               ovar[:] = ivar.mean(axis=-1)
           else:
               ovar[:] = ivar[:]

   # copy dynamic variables (time-dependent) making latitudinal means
   for tt in range(ntime):
       for ivar in fi.variables.values():
           if vtime in ivar.dimensions:
               ovar = fo.variables[ivar.name]
               if vlon in ivar.dimensions:
                   ovar[tt, ...] = ivar[tt, ...].mean(axis=-1)
               else:
                   if ivar.ndim == 1:
                       ovar[tt] = ivar[tt]
                   else:
                       ovar[tt, ...] = ivar[tt, ...]

   # set new variable
   shape1 = fi.variables['var1'].shape
   shape1 = shape1[:-1]
   var2[:] = np.arange(np.prod(shape1)).reshape(shape1)

   # finish
   fi.close()
   fo.close()

:copyright: Copyright 2020-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

Subpackages
===========
.. autosummary::
   netcdfio

"""
from .netcdfio import copy_dimensions, copy_file
from .netcdfio import copy_global_attributes, copy_variables
from .netcdfio import create_new_variable, create_variables
from .netcdfio import get_fill_value_for_dtype, set_output_filename

__all__ = ['copy_dimensions', 'copy_file', 'copy_global_attributes',
           'copy_variables', 'create_new_variable', 'create_variables',
           'get_fill_value_for_dtype', 'set_output_filename']



#!/usr/bin/env python
"""
Calculate daily, monthly, yearly, etc. means

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2013-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   means

History
    * Written Jul 2013 by Matthias Cuntz (mc (at) macu (dot) de)
    * Added meanday, Jul 2013, Matthias Cuntz
    * Added max and min, Apr 2014, Matthias Cuntz
    * Added onlydat and meanmonth, Jun 2015, Matthias Cuntz
    * Added seasonal, Oct 2018, Matthias Cuntz
    * Bug in minute averaging: compared minutes to hours,
      Jun 2019, Matthias Cuntz
    * Added half_hour, inspired by Robin Leucht for UFZ level1 data;
      did not take retrospective keyword but added Note for this case,
      Jun 2019, Matthias Cuntz
    * Use pyjams datetime, Jul 2022, Matthias Cuntz
    * More possible date formats, Jul 2022, Matthias Cuntz
    * Rename seasonal to seasonalday, Jul 2022, Matthias Cuntz
    * Add seasonalmonth, Jul 2022, Matthias Cuntz
    * Add seasonalmeanday, Jul 2022, Matthias Cuntz

"""
import datetime as dt
import cftime as cf
import numpy as np
from .class_datetime import datetime
from .class_datetime import date2num, num2date
from .helper import input2array, array2input
# from pyjams import datetime
# from pyjams import date2num, num2date
# from pyjams.helper import input2array, array2input


__all__ = ['means']


def _to_tuple(dt):
    """
    Turn a datetime instance into a tuple of integers. Elements go
    in the order of decreasing significance, making it easy to compare
    datetime instances. Parts of the state that don't affect ordering
    are omitted.

    """
    return (dt.year, dt.month, dt.day, dt.hour, dt.minute,
            dt.second, dt.microsecond)


def _output_date(dout, ical, date0, idatetime, idkwargs):
    """
    Conversion of output dates to input date format

    """
    aout = num2date(dout, calendar=ical)
    if idatetime:
        if isinstance(date0, dt.datetime):
            ddout = [ idatetime(*_to_tuple(aa)) for aa in aout ]
        else:
            ddout = [ idatetime(*_to_tuple(aa), calendar=date0.calendar)
                      for aa in aout ]
    elif isinstance(date0, str):
        oform = idkwargs['format']
        if not oform:
            oform = None
        ddout = [ aa.strftime(oform) for aa in aout ]
    else:
        okwargs = {'units': idkwargs['units'],
                   'calendar': idkwargs['calendar'],
                   'has_year_zero': idkwargs['has_year_zero']}
        ddout = date2num(aout, **okwargs)

    return ddout


def means(date, dat,
          year=False, month=False, day=False, hour=False, half_hour=False,
          minute=False,
          meanday=False, meanmonth=False, seasonalday=False,
          seasonalmonth=False, seasonalmeanday=False,
          sum=False, max=False, min=False,
          onlydat=False, **kwargs):
    """
    Calculate daily, monthly, yearly, etc. means of data

    dat will be averaged over columns, i.e. assuming that the first
    dimension is the time dimension.

    If no keyword argument is given, the mean will be over the whole
    first column.

    date can be either string or datetime object of Python, numpy, cftime, or
    pyjams, which will be interpreted by pyjams.date2num; or date can be a
    numeric value, which will be interpreted by pyjams.num2date; any extra
    keyword argument will be passed to the appropriate datetime routine.
    Returned dates will be in same format as incoming dates.

    Parameters
    ----------
    date : 1D array with dates
        Can either be strings or *datetime* objects of *Python*, *cftime* or
        *pyjams*, which will be interpreted by *pyjams.date2num*, or it
        can be a numeric value, which will be interpreted by *pyjams.num2date*.
        date should be centred on the input time period, i.e. half-hourly data
        should have dates at 15 min and 45 min of each hour, for example.
    dat : ND (masked-)array
    year : bool, optional
        If True, return annual means
    month : bool, optional
        If True, return monthly means
    day : bool, optional
        If True, return daily means
    hour : bool, optional
        If True, return hourly means
    half_hour : bool, optional
        If True, return half-hourly means
    minute : bool, optional
        If True, return minute means. Note that this can take very long.
    meanday : bool, optional
        If True, return one mean daily cycle
    meanmonth : bool, optional
        If True, return mean monthly cycle
    seasonalday : bool, optional
        If True, return daily means seasonal cycle
    seasonalmonth : bool, optional
        Same as meanmonth. If True, return mean monthly cycle
    seasonalmeanday : bool, optional
        If True, return one mean daily cycle per month
    sum : bool, optional
        If True, calculate sums instead of means
    max : bool, optional
        If True, calculate maxima instead of means
    min : bool, optional
        If True, calculate minima instead of means
    onlydat : bool, optional
        If True, return only meandat, else return [outdate, meandat]
    **kwargs : dict, optional
        Any other keyword argument will be passed to either *pyjams.date2num*
        or *pyjams.num2date* depending on type of *date*

    Returns
    -------
    outdate, meandat
       centred date on year, month, day, etc.; and
       averaged data in 1st dimension

    Notes
    -----
    *outdate* is same format as *date* input. Output *meandat* is scalar,
    numpy array, or masked array if needed.

    Returns centred [dates, averages]:
        * Yearly dates are centred on June 15, 12:00 h
        * Monthly dates are centred on 15th, 12:00 h
        * Daily dates are centred on 12:00 h
        * Hourly dates are centred on 30 min
        * Half hourly dates are centred on 15 and 45 min
        * Minute dates are centred on 30 sec
        * Mean daily dates are centred on 30 min of January 01 of the first
          year
        * Mean monthly (seasonal monthly) dates are centred on the 15th of
          each month at 12:00 h of the first year
        * Seasonal daily dates are centred on 12:00 h of each day of the
          first (leap) year
        * Seasonal mean daily dates are centred on 30 min of each first day of
          each month of the first year

    If meanday/seasonal/meanday/seasonalday/meanmonth/seasonalmonth==True,
    then all hours/months/days will be written;
    as a masked-array if hours/days/months are missing.

    If seasonalday==True, input data has to be spaced <= days, otherwise
    consider meanmonth/seasonalmonth.

    If input date represent the beginning/end of the time step, the user should
    add/remove half a time step before using the routine. Otherwise the routine
    would have to guess the time step, which is error prone.
    For example, remove 15 minutes by subtracting
    `datetime.timedelta(minutes=15)` from datetime, `15./(24.*60.)` from Julian
    dates, or `pyjams.date2num('2010-01-01 12:15', units, calendar) -
    pyjams.date2num('2010-01-01 12:00', units, calendar)` from any *calendar*
    and *units* understood by *pyjams.datetime* or *cftime.datetime*.

    Examples
    --------
    >>> dates = ['01.01.1990 12:00:00', '01.02.1990 12:00:00',
    ...          '01.03.1990 12:00:00', '01.01.1990 12:10:00',
    ...          '01.01.1990 13:00:00', '01.01.1990 14:00:00']
    >>> jdates = date2num(dates, calendar='decimal')
    >>> x = np.arange(len(jdates)) + 1.
    >>> odates, xout = means(jdates, x, calendar='decimal')
    >>> oodates = num2date(odates, calendar='decimal')
    >>> oodates = oodates.strftime('%Y-%m-%d %H:%M:%S')
    >>> print(oodates, xout)
    1990-01-31 00:00:00 3.5

    >>> odates, xout = means(jdates, x, year=True, calendar='decimal')
    >>> print(xout)
    [3.5]

    >>> odates, xout = means(dates, x, month=True)
    >>> print(xout)
    [4. 2. 3.]

    >>> odates, xout = means(dates, x, day=True)
    >>> print(xout)
    [4. 2. 3.]

    >>> odates, xout = means(dates, x, hour=True)
    >>> print(xout)
    [2.5 5.  6.  2.  3. ]

    >>> odates, xout = means(jdates, x, half_hour=True, calendar='decimal')
    >>> print(xout)
    [2.5 5.  6.  2.  3. ]

    >>> odates, xout = means(jdates, x, meanday=True, calendar='decimal')
    >>> print(xout[10:16])
    [-- -- 2.5 5.0 6.0 --]

    >>> odates, xout = means(jdates, x, meanmonth=True, calendar='decimal')
    >>> print(xout[0:5])
    [4.0 2.0 3.0 --  --]

    >>> odates, xout = means(jdates, x, seasonalday=True, calendar='decimal')
    >>> print(xout[0:5])
    [4.0 -- -- -- --]

    >>> odates, xout = means(jdates, x, seasonalmonth=True, calendar='decimal')
    >>> print(xout[0:5])
    [4.0 2.0 3.0 --  --]

    >>> print(means(jdates, x, month=True, onlydat=True, calendar='decimal'))
    [4. 2. 3.]

    Masked arrays

    >>> x = np.ma.array(x, mask=np.zeros(x.size, dtype=bool))
    >>> x.mask[0] = True
    >>> odates, xout = means(jdates, x, calendar='decimal')
    >>> print(np.around(odates, 1), xout)
    1990.1 4.0

    >>> odates, xout = means(dates, x, year=True)
    >>> print(xout)
    [4.0]

    >>> odates, xout = means(dates, x, month=True)
    >>> print(xout)
    [5.0 2.0 3.0]

    >>> odates, xout = means(jdates, x, day=True, calendar='decimal')
    >>> print(xout)
    [5.0 2.0 3.0]

    sum

    >>> odates, xout = means(jdates, x, sum=True, calendar='decimal')
    >>> print(np.around(odates, 1), xout)
    1990.1 20.0
    >>> odates, xout = means(jdates, x, year=True, sum=True,
    ...                      calendar='decimal')
    >>> print(xout)
    [20.0]
    >>> odates, xout = means(dates, x, month=True, sum=True)
    >>> print(xout)
    [15.0 2.0 3.0]
    >>> odates, xout = means(jdates, x, day=True, sum=True, calendar='decimal')
    >>> print(xout)
    [15.0 2.0 3.0]

    max

    >>> odates, xout = means(jdates, x, max=True, calendar='decimal')
    >>> print(np.around(odates, 1), xout)
    1990.1 6.0
    >>> odates, xout = means(dates, x, year=True, max=True)
    >>> print(xout)
    [6.0]
    >>> odates, xout = means(jdates, x, month=True, max=True,
    ...                      calendar='decimal')
    >>> print(xout)
    [6.0 2.0 3.0]
    >>> odates, xout = means(jdates, x, day=True, max=True, calendar='decimal')
    >>> print(xout)
    [6.0 2.0 3.0]

    min

    >>> odates, xout = means(jdates, x, min=True, calendar='decimal')
    >>> print(np.around(odates, 1), xout)
    1990.1 2.0
    >>> odates, xout = means(jdates, x, year=True, min=True,
    ...                      calendar='decimal')
    >>> print(xout)
    [2.0]
    >>> odates, xout = means(dates, x, month=True, min=True)
    >>> print(xout)
    [4.0 2.0 3.0]
    >>> odates, xout = means(jdates, x, day=True, min=True, calendar='decimal')
    >>> print(xout)
    [4.0 2.0 3.0]

    2D and masked arrays

    >>> x  = np.repeat(x,2).reshape((x.size,2))
    >>> odates, xout = means(jdates, x, calendar='decimal')
    >>> print(np.around(odates, 1), xout)
    1990.1 [4.0 4.0]

    >>> odates, xout = means(jdates, x, year=True, calendar='decimal')
    >>> print(xout)
    [[4.0 4.0]]

    >>> odates, xout = means(dates, x, month=True)
    >>> print(xout)
    [[5.0 5.0]
     [2.0 2.0]
     [3.0 3.0]]

    >>> odates, xout = means(jdates, x, day=True, calendar='decimal')
    >>> print(xout)
    [[5.0 5.0]
     [2.0 2.0]
     [3.0 3.0]]

    """
    # Constants
    myundef = 9e33
    ismasked = isinstance(dat, np.ma.MaskedArray)

    # Assure array
    date = np.array(date)
    if not ismasked:
        dat = np.array(dat)

    # Assure ND-array
    isone = False
    if dat.ndim == 1:
        isone = True
        dat = dat[:, np.newaxis]

    # Check options
    if seasonalmonth:
        meanmonth = True
    allopts = (day + month + year + hour + half_hour + minute + meanday +
               meanmonth + seasonalday + seasonalmeanday)
    assert allopts <= 1, (
        "only one averaging option day, month, year, etc. possible")

    # Check aggregation
    allaggs = sum + max + min
    assert allaggs <= 1, "only one aggregation option sum, min, max possible"

    # transform date
    iform = '%Y-%m-%d %H:%M:%S'
    ical = 'decimal'
    idatetime = ''
    idkwargs = {}
    date0 = date[0]
    if isinstance(date0, (dt.datetime, cf.datetime, datetime)):
        if isinstance(date0, dt.datetime):
            idatetime = dt.datetime
        elif isinstance(date0, cf.datetime):
            idatetime = cf.datetime
        elif isinstance(date0, datetime):
            idatetime = datetime
        # datetime
        adate = [ dd.strftime(iform) for dd in date ]
        idate = date2num(adate, calendar=ical)
    elif isinstance(date0, str):
        # str
        idkwargs = {'format': '', 'timesep': ' ', 'fr': False}
        for kk in kwargs:
            if kk in idkwargs:
                idkwargs.update({kk: kwargs[kk]})
        idate = date2num(date, calendar=ical, **idkwargs)
    else:
        # num - should throw an error if other
        idkwargs = {'units': '', 'calendar': 'standard',
                    'only_use_cftime_datetimes': True,
                    'only_use_python_datetimes': False,
                    'has_year_zero': None,
                    'format': '', 'return_arrays': False}
        for kk in kwargs:
            if kk in idkwargs:
                idkwargs.update({kk: kwargs[kk]})
        ddate = num2date(date, **idkwargs)
        adate = [ dd.strftime(iform) for dd in ddate ]
        idate = date2num(adate, calendar=ical)
    idate = input2array(idate, default=1.)

    # select operator
    if sum:
        meanop = np.ma.sum
    elif max:
        meanop = np.ma.amax
    elif min:
        meanop = np.ma.amin
    else:
        meanop = np.ma.mean
    if ismasked:
        ones = np.ma.ones
    else:
        ones = np.ones

    # average 1st dim
    if allopts == 0:
        dout = [0.5 * (idate.max() + idate.min())]
        ddout = _output_date(dout, ical, date0, idatetime, idkwargs)
        ddout = ddout[0]
        out = meanop(dat, 0)
        if isone:
            if onlydat:
                return out[0]
            else:
                return ddout, out[0]
        else:
            if onlydat:
                return out
            else:
                return ddout, out

    # if allopts > 0
    # yr, mo, dy, hr, mn, sc = dec2date(date)
    yr, mo, dy, hr, mn, sc, ms = num2date(
        idate, calendar='decimal', return_arrays=True)

    # year
    if year:
        yrs = np.unique(yr)
        nout = yrs.size
        dout = ['9999-01-01 00:00:00'] * nout
        out = ones([nout] + list(dat.shape[1:])) * myundef
        zahl = 0
        for i in yrs:
            ii = np.where(yr == i)[0]
            if np.size(ii) > 0:
                dout[zahl] = ('{:04d}-{:02d}-{:02d}'
                              ' {:02d}:{:02d}:{:02d}'.format(
                                  i, 6, 15, 12, 0, 0))
                out[zahl, :] = meanop(dat[ii, :], 0)
                zahl += 1

    # month
    if month:
        yrs = np.unique(yr)
        mos = np.unique(mo)
        nout = yrs.size * mos.size
        dout = ['9999-01-01 00:00:00'] * nout
        out = ones([nout] + list(dat.shape[1:])) * myundef
        zahl = 0
        for i in yrs:
            for j in mos:
                ii = np.where((yr == i) & (mo == j))[0]
                if np.size(ii) > 0:
                    dout[zahl] = ('{:04d}-{:02d}-{:02d}'
                                  ' {:02d}:{:02d}:{:02d}'.format(
                                      i, j, 15, 12, 0, 0))
                    out[zahl, :] = meanop(dat[ii, :], 0)
                    zahl += 1

    # day
    if day:
        yrs = np.unique(yr)
        mos = np.unique(mo)
        dys = np.unique(dy)
        nout = yrs.size * mos.size * dys.size
        dout = ['9999-01-01 00:00:00'] * nout
        out = ones([nout] + list(dat.shape[1:])) * myundef
        zahl = 0
        for i in yrs:
            for j in mos:
                for k in dys:
                    ii = np.where((yr == i) & (mo == j) & (dy == k))[0]
                    if np.size(ii) > 0:
                        dout[zahl] = ('{:04d}-{:02d}-{:02d}'
                                      ' {:02d}:{:02d}:{:02d}'.format(
                                          i, j, k, 12, 0, 0))
                        out[zahl, :] = meanop(dat[ii, :], 0)
                        zahl += 1

    # hour
    if hour:
        yrs = np.unique(yr)
        mos = np.unique(mo)
        dys = np.unique(dy)
        hrs = np.unique(hr)
        nout = yrs.size * mos.size * dys.size * hrs.size
        dout = ['9999-01-01 00:00:00'] * nout
        out = ones([nout] + list(dat.shape[1:])) * myundef
        zahl = 0
        for i in yrs:
            for j in mos:
                for k in dys:
                    for l in hrs:
                        ii = np.where((yr == i) & (mo == j) &
                                      (dy == k) & (hr == l))[0]
                        if np.size(ii) > 0:
                            dout[zahl] = ('{:04d}-{:02d}-{:02d}'
                                          ' {:02d}:{:02d}:{:02d}'.format(
                                              i, j, k, l, 30, 0))
                            out[zahl, :] = meanop(dat[ii, :], 0)
                            zahl += 1
    # half_hour
    if half_hour:
        yrs = np.unique(yr)
        mos = np.unique(mo)
        dys = np.unique(dy)
        hrs = np.unique(hr)
        nout = yrs.size * mos.size * dys.size * hrs.size * 2
        dout = ['9999-01-01 00:00:00'] * nout
        out = ones([nout] + list(dat.shape[1:])) * myundef
        zahl = 0
        for i in yrs:
            for j in mos:
                for k in dys:
                    for l in hrs:
                        for m in range(2):
                            ii = np.where((yr == i) & (mo == j) &
                                          (dy == k) & (hr == l) &
                                          ((mn // 30) == m))[0]
                            if np.size(ii) > 0:
                                dout[zahl] = ('{:04d}-{:02d}-{:02d}'
                                              ' {:02d}:{:02d}:{:02d}'.format(
                                                  i, j, k, l, 15 + m * 30, 0))
                                out[zahl, :] = meanop(dat[ii, :], 0)
                                zahl += 1

    # minute
    if minute:
        yrs = np.unique(yr)
        mos = np.unique(mo)
        dys = np.unique(dy)
        hrs = np.unique(hr)
        mns = np.unique(mn)
        nout = yrs.size * mos.size * dys.size * hrs.size * mns.size
        dout = ['9999-01-01 00:00:00'] * nout
        out = ones([nout] + list(dat.shape[1:])) * myundef
        zahl = 0
        for i in yrs:
            for j in mos:
                for k in dys:
                    for l in hrs:
                        for m in mns:
                            ii = np.where((yr == i) & (mo == j) &
                                          (dy == k) & (hr == l) &
                                          (mn == m))[0]
                            if np.size(ii) > 0:
                                dout[zahl] = ('{:04d}-{:02d}-{:02d}'
                                              ' {:02d}:{:02d}:{:02d}'.format(
                                                  i, j, k, l, m, 30))
                                out[zahl, :] = meanop(dat[ii, :], 0)
                                zahl += 1

    # Mean daily
    if meanday:
        nout = 24
        hrs = range(nout)
        dout = ['9999-01-01 00:00:00'] * nout
        out = ones([nout] + list(dat.shape[1:])) * myundef
        zahl = 0
        for i in hrs:
            dout[zahl] = '{:04d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}'.format(
                yr[0], 1, 1, i, 30, 0)
            ii = np.where(hr == i)[0]
            if np.size(ii) > 0:
                out[zahl, :] = meanop(dat[ii, :], 0)
            zahl += 1
        if np.any(out == myundef):
            out = np.ma.array(out, mask=(out == myundef), keep_mask=True)

    # Mean monthly = seasonal monthly
    if meanmonth:  # seasonalmonth
        nout = 12
        mos = range(1, nout + 1)
        dout = ['9999-01-01 00:00:00'] * nout
        out = ones([nout] + list(dat.shape[1:])) * myundef
        zahl = 0
        for i in mos:
            dout[zahl] = '{:04d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}'.format(
                yr[0], i, 15, 12, 0, 0)
            ii = np.where(mo == i)[0]
            if np.size(ii) > 0:
                out[zahl, :] = meanop(dat[ii, :], 0)
            zahl += 1
        if np.any(out == myundef):
            out = np.ma.array(out, mask=(out == myundef), keep_mask=True)

    # Seasonalday
    if seasonalday:
        dim = np.array([[-9, 31, 28, 31, 30, 31, 30, 31, 31,
                         30, 31, 30, 31],
                        [-9, 31, 29, 31, 30, 31, 30, 31, 31,
                         30, 31, 30, 31]])
        leaps = (((yr % 4) == 0) & ((yr % 100) != 0)) | ((yr % 400) == 0)
        leap = np.any(leaps)
        ileap = int(leap)
        if leap:
            iileap = np.argmax(leaps)
            nout = 366
        else:
            iileap = 0
            nout = 365
        dys = range(1, nout + 1)
        dout = ['9999-01-01 00:00:00'] * nout
        out = ones([nout] + list(dat.shape[1:])) * myundef
        zahl = 0
        for i in range(1, 13):
            for j in range(1, dim[ileap, i]+1):
                dout[zahl] = ('{:04d}-{:02d}-{:02d}'
                              ' {:02d}:{:02d}:{:02d}'.format(
                                  yr[iileap], i, j, 12, 0, 0))
                ii = np.where((mo == i) & (dy == j))[0]
                if np.size(ii) > 0:
                    out[zahl, :] = meanop(dat[ii, :], 0)
                zahl += 1
        if np.any(out == myundef):
            out = np.ma.array(out, mask=(out == myundef), keep_mask=True)

    # seasonal mean daily
    if seasonalmeanday:
        nmonth = 12
        mos = range(1, nmonth+1)
        nhour = 24
        hrs = range(nhour)
        nout = nmonth * nhour
        dout = ['9999-01-01 00:00:00'] * nout
        out = ones([nout] + list(dat.shape[1:])) * myundef
        zahl = 0
        for i in mos:
            for j in hrs:
                dout[zahl] = '{:04d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}'.format(
                    yr[0], i, 1, j, 30, 0)
                ii = np.where((mo == i) & (hr == j))[0]
                if np.size(ii) > 0:
                    out[zahl, :] = meanop(dat[ii, :], 0)
                zahl += 1
        if np.any(out == myundef):
            out = np.ma.array(out, mask=(out == myundef), keep_mask=True)

    # remove mask if possible
    if np.ma.count_masked(out) == 0:
        out = np.array(out)

    # decimal date
    dout = np.array(date2num(dout, calendar=ical))

    ii = np.where(dout != 9999.)[0]
    if len(ii) > 0:
        dout = dout[ii]
        out = out[ii, :]
    else:  # pragma: no cover
        # should not end up here
        print(dout)
        raise ValueError("all date values undefined after mean.")

    # decimal date to initial date format
    ddout = _output_date(dout, ical, date0, idatetime, idkwargs)
    ddout = array2input(ddout, date)

    if isone:
        if onlydat:
            return out[:, 0]
        else:
            return ddout, out[:, 0]
    else:
        if onlydat:
            return out
        else:
            return ddout, out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # from jams.autostring import astr
    # from jams.date2dec import date2dec
    # from jams.dec2date import dec2date
    # dates = ['01.01.1990 12:00:00', '01.02.1990 12:00:00',
    #          '01.03.1990 12:00:00', '01.01.1990 12:10:00',
    #          '01.01.1990 13:00:00', '01.01.1990 14:00:00']
    # jdates = date2dec(ascii=dates)
    # x = np.arange(len(jdates)) + 1.
    # odates, xout = means(jdates, x)
    # oodates = dec2date(odates, en=True)
    # print(oodates, xout)
    # # 1990-01-31 00:00:00 3.5

    # odates, xout = means(jdates, x, year=True)
    # print(xout)
    # # [3.5]

    # odates, xout = means(jdates, x, month=True)
    # print(xout)
    # # [4. 2. 3.]

    # Odates, xout = means(jdates, x, day=True)
    # print(xout)
    # # [4. 2. 3.]

    # odates, xout = means(jdates, x, hour=True)
    # print(xout)
    # # [2.5 5.  6.  2.  3. ]

    # odates, xout = means(jdates, x, half_hour=True)
    # print(xout)
    # # ['2.500' '5.000' '6.000' '2.000' '3.000']

    # odates, xout = means(jdates, x, meanday=True)
    # print(astr(xout[10:16], 3, pp=True))
    # # ['--   ' '--   ' '2.500' '5.000' '6.000' '--   ']

    # odates, xout = means(jdates, x, meanmonth=True)
    # print(astr(xout[0:5], 3, pp=True))
    # # ['4.000' '2.000' '3.000' '--   ' '--   ']

    # odates, xout = means(jdates, x, seasonalday=True)
    # print(astr(xout[0:5], 3, pp=True))
    # # ['4.000' '--   ' '--   ' '--   ' '--   ']

    # print(astr(means(jdates, x, month=True, onlydat=True), 3, pp=True))
    # # ['4.000' '2.000' '3.000']

    # # Masked arrays
    # x = np.ma.array(x, mask=np.zeros(x.size, dtype=bool))
    # x.mask[0] = True
    # odates, xout = means(jdates, x)
    # print(astr(odates, 3, pp=True), astr(xout, 3, pp=True))
    # # 2.448e+06 4.000

    # odates, xout = means(jdates, x, year=True)
    # print(astr(xout, 3, pp=True))
    # # ['4.000']

    # odates, xout = means(jdates, x, month=True)
    # print(astr(xout, 3, pp=True))
    # # ['5.000' '2.000' '3.000']

    # odates, xout = means(jdates, x, day=True)
    # print(astr(xout, 3, pp=True))
    # # ['5.000' '2.000' '3.000']

    # # sum
    # odates, xout = means(jdates, x, sum=True)
    # print(astr(odates, 3, pp=True), astr(xout, 3, pp=True))
    # # 2.448e+06 20.000
    # odates, xout = means(jdates, x, year=True, sum=True)
    # print(astr(xout, 3, pp=True))
    # # ['20.000']
    # odates, xout = means(jdates, x, month=True, sum=True)
    # print(astr(xout, 3, pp=True))
    # # ['15.000' ' 2.000' ' 3.000']
    # odates, xout = means(jdates, x, day=True, sum=True)
    # print(astr(xout, 3, pp=True))
    # # ['15.000' ' 2.000' ' 3.000']

    # # max
    # odates, xout = means(jdates, x, max=True)
    # print(astr(odates, 3, pp=True), astr(xout, 3, pp=True))
    # # 2.448e+06 6.000
    # odates, xout = means(jdates, x, year=True, max=True)
    # print(astr(xout, 3, pp=True))
    # # ['6.000']
    # odates, xout = means(jdates, x, month=True, max=True)
    # print(astr(xout, 3, pp=True))
    # # ['6.000' '2.000' '3.000']
    # odates, xout = means(jdates, x, day=True, max=True)
    # print(astr(xout, 3, pp=True))
    # # ['6.000' '2.000' '3.000']

    # # min
    # odates, xout = means(jdates, x, min=True)
    # print(astr(odates, 3, pp=True), astr(xout, 3, pp=True))
    # # 2.448e+06 2.000
    # odates, xout = means(jdates, x, year=True, min=True)
    # print(astr(xout, 3, pp=True))
    # # ['2.000']
    # odates, xout = means(jdates, x, month=True, min=True)
    # print(astr(xout, 3, pp=True))
    # # ['4.000' '2.000' '3.000']
    # odates, xout = means(jdates, x, day=True, min=True)
    # print(astr(xout, 3, pp=True))
    # # ['4.000' '2.000' '3.000']

    # # 2D and masked arrays
    # x  = np.repeat(x,2).reshape((x.size,2))
    # odates, xout = means(jdates, x)
    # print(astr(odates, 3, pp=True), astr(xout, 3, pp=True))
    # # 2.448e+06 ['4.000' '4.000']

    # odates, xout = means(jdates, x, year=True)
    # print(astr(xout, 3, pp=True))
    # # [['4.000' '4.000']]

    # odates, xout = means(jdates, x, month=True)
    # print(astr(xout, 3, pp=True))
    # # [['5.000' '5.000']
    # #  ['2.000' '2.000']
    # #  ['3.000' '3.000']]

    # odates, xout = means(jdates, x, day=True)
    # print(astr(xout, 3, pp=True))
    # # [['5.000' '5.000']
    # #  ['2.000' '2.000']
    # #  ['3.000' '3.000']]
